#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Class
2022-02-01 Félix Léger - felixleger@gmail.com & sakic@ipgp.fr
"""

import os
import re
import string
from datetime import datetime, timedelta
from io import StringIO
from pathlib import Path

import hatanaka

# import matplotlib.pyplot as plt
import numpy as np

import rinexmod.logger as rimo_log

logger = rimo_log.logger_define("INFO")


# logger = logging.getLogger("rinexmod_api")

# from rinexmod import rinexmod_api as rimo_api
# logger = rimo_api.logger_define('INFO')


# *****************************************************************************
# class definition
class RinexFile:
    """
    Will store a compressed rinex file content in a file-like list of strings
    using the hatanaka library.
    Will then provide methods to modifiy the file's header.
    The method to get the sample rate will read the whole file and will set as
    unknow sample rate files that have more than 10% of non-nominal sample rate.
    The method to get the file duration is based on reading the file name and
    not the data.
    A method to write the file in selected compression is also available.
    The get_header method permits to have a printable string of all file's metadata.
    """

    def __init__(self, rinex_input, force_rnx_load=False):

        ##### the RINEX input is a file, thus a string or a Path is given
        if type(rinex_input) in (str, Path):
            self.source_from_file = True
            self.path = rinex_input.strip()
            self.path_output = ""
            self.content_input = self.path
        ##### the RINEX input is directely data, in a StringIO
        elif type(rinex_input) is StringIO:
            self.source_from_file = False
            self.path = ""
            self.path_output = ""
            self.content_input = bytes(
                rinex_input.read(), "utf-8"
            )  ## ready-to-read for hatanaka
        else:
            logger.error("input rinex_input is not str, Path, or StringIO")

        self.rinex_data, self.status = self._load_rinexdata(
            force_rnx_load=force_rnx_load
        )
        self.size = self.get_size()
        self.name_conv = self.get_naming_convention()
        self.compression, self.hatanka_input = self.get_compression()
        self.filename = self.get_filename()
        self.data_source = self.get_data_source()

        #### site is an internal attribute, always 9char
        # (filled with 00XXX in nothing else is provided)
        # it is accessible wit get_site()
        self._site = self.get_site_from_filename(lower_case=False, only_4char=False)

        self.version, self.version_float = self.get_version()
        self.start_date, self.end_date = self.get_dates()
        self.sample_rate_string, self.sample_rate_numeric = self.get_sample_rate(
            plot=False
        )
        # self.file_period, self.session = self.get_file_period_from_filename()
        self.file_period, self.session = self.get_file_period_from_data(
            inplace_set=False
        )
        ### NB: here, when we load the RINEX, we remain tolerant for the file period!!

    def __repr__(self):
        """
        Defines a representation method for the rinex file object. Will print the
        filename, the site, the start and end date, the sample rate, the file period,
        the rinex version, the data source, the compression type, the size of the file,
        and the status of the file.
        """
        if not self.rinex_data:
            return ""

        return (
            f"RinexFile: {self.filename}\n"
            f"Site: {self.get_site()}\n"
            f"Start date: {self.start_date}\n"
            f"End date: {self.end_date}\n"
            f"Sample rate: {self.sample_rate_string}\n"
            f"File period: {self.file_period}\n"
            f"Rinex version: {self.version}\n"
            f"Data source: {self.data_source}\n"
            f"Compression: {self.compression}\n"
            f"Size: {self.size} bytes\n"
            f"Status: {self.status}"
        )

    def __str__(self):
        """
        Defines a print method for the rinex file object. Will print all the
        header, plus 20 lines of data, plus the number of not printed lines.
        """
        if not self.rinex_data:
            return ""

        # We get header
        end_of_header_idx = search_idx_value(self.rinex_data, "END OF HEADER") + 1
        str_rinex_file = self.rinex_data[0:end_of_header_idx]
        # We add 20 lines of data
        str_rinex_file.extend(
            self.rinex_data[end_of_header_idx : end_of_header_idx + 20]
        )
        # We add a line that contains the number of lines not to be printed
        length = len(self.rinex_data) - len(str_rinex_file) - 20
        cut_lines_rep = (
            " " * 20
            + "["
            + "{} lines hidden".format(length).center(40, ".")
            + "]"
            + " " * 20
        )
        str_rinex_file.append(cut_lines_rep)

        return "\n".join(str_rinex_file)

    # *****************************************************************************
    #### Main methods

    def get_header(self):
        """
        Returns a printable, with carriage-return, string of metadata lines from
        the header, and a python dict of the same information.
        """

        # if self.status not in [0, 2]:
        #    return None

        if self.status:
            return None

        head_raw = {}

        head_labels = [
            "RINEX VERSION / TYPE",
            "MARKER NAME",
            "MARKER NUMBER",
            "OBSERVER / AGENCY",
            "REC # / TYPE / VERS",
            "ANT # / TYPE",
            "APPROX POSITION XYZ",
            "ANTENNA: DELTA H/E/N",
            "TIME OF FIRST OBS",
        ]

        missing_head_labels = []
        for kw in head_labels:
            for line in self.rinex_data:
                if kw in line and not kw in head_raw.keys():
                    head_raw[kw] = line
                    break
                if "END OF HEADER" in line:
                    missing_head_labels.append(kw)
                    break

        for miss_lab in missing_head_labels:
            logger.warning("%s label is missing in %s header", miss_lab, self.filename)
            head_raw[miss_lab] = ""

        ### previous case with just 'MARKER NUMBER', too weak hardcoded solution
        # if not 'MARKER NUMBER' in metadata:
        #    metadata['MARKER NUMBER'] = ''

        head_out = {
            "File": self.path,
            "File size (bytes)": self.size,
            "Rinex version": head_raw["RINEX VERSION / TYPE"][0:9].strip(),
            "Sample rate": self.sample_rate_string,
            "File period": self.file_period,
            "Observable type": head_raw["RINEX VERSION / TYPE"][40:60].strip(),
            "Marker name": head_raw["MARKER NAME"][0:60].strip(),
            "Marker number": head_raw["MARKER NUMBER"][0:60].strip(),
            "Operator": head_raw["OBSERVER / AGENCY"][0:20].strip(),
            "Agency": head_raw["OBSERVER / AGENCY"][20:40].strip(),
            "Receiver serial": head_raw["REC # / TYPE / VERS"][0:20].strip(),
            "Receiver type": head_raw["REC # / TYPE / VERS"][20:40].strip(),
            "Receiver firmware version": head_raw["REC # / TYPE / VERS"][40:60].strip(),
            "Antenna serial": head_raw["ANT # / TYPE"][0:20].strip(),
            "Antenna type": head_raw["ANT # / TYPE"][20:40].strip(),
            "Antenna position (XYZ)": "{:14} {:14} {:14}".format(
                head_raw["APPROX POSITION XYZ"][0:14].strip(),
                head_raw["APPROX POSITION XYZ"][14:28].strip(),
                head_raw["APPROX POSITION XYZ"][28:42].strip(),
            ),
            "Antenna delta (H/E/N)": "{:14} {:14} {:14}".format(
                head_raw["ANTENNA: DELTA H/E/N"][0:14].strip(),
                head_raw["ANTENNA: DELTA H/E/N"][14:28].strip(),
                head_raw["ANTENNA: DELTA H/E/N"][28:42].strip(),
            ),
            "Start date and time": self.start_date,
            "Final date and time": self.end_date,
        }

        head_str = "\n".join(["{:29} : {}".format(k, v) for k, v in head_out.items()])

        head_str = "\n" + head_str + "\n"

        return head_str, head_out

    def get_site(self, lower_case=True, only_4char=False, no_country_then_4char=False):
        """
        Get the site of the Rinex Object
        if no_country_then_4char = True, the return site code
        falls back to a 4-char code, if no monum/country is provided
        (i.e. the 9char code ends with 00XXX)
        if no_country_then_4char = False, 00XXX will remain
        """

        site_out = self._site

        if lower_case:
            site_out = site_out.lower()
        else:
            site_out = site_out.upper()

        if only_4char or (no_country_then_4char and site_out[-3:] == "XXX"):
            site_out = site_out[0:4]

        return site_out

    def set_site(self, site_4or9char, monum=None, country=None):
        """
        set site for a Rinex Object
        monum and country overrides the ones given if site_4or9char
        is 9 char. long
        """

        if len(site_4or9char) not in (4, 9):
            logger.error("given site is not 4 nor 9 char. long!")
            return None

        if len(site_4or9char) == 4:
            if not monum:
                monum = "00"
            if not country:
                country = "XXX"
            self._site = site_4or9char + monum + country
        else:
            if not monum:
                monum = site_4or9char[4:6]
            if not country:
                country = site_4or9char[6:]
            self._site = site_4or9char + monum + country

        return None

    def get_longname(
        self,
        data_source="R",
        obs_type="O",
        ext="auto",
        compression="auto",
        filename_style="basic",
        inplace_set=False,
    ):
        """
        Generate the long RINEX filename.
        Can be stored directly as filename attribute with inplace=True.

        Parameters
        ----------
        data_source : str, optional
            The data source identifier. Default is 'R'.
        obs_type : str, optional
            The observation type. Default is 'O'.
        ext : str, optional
            The file extension. 'auto' to determine based on the RinexFile attribute,
            or manually specify 'rnx' or 'crx'. Default is 'auto'.
        compression : str, optional
            The compression type. 'auto' to determine based on the RinexFile attribute,
            or manually specify 'Z', 'gz', etc. Default is 'auto'.
        filename_style : str, optional
            The style of the filename.
             Possible values are 'basic', 'flex', 'exact'.
             See the rinexmod main function's docstring/help for more details.
             Default is 'basic'.
        inplace_set : bool, optional
            If True, stores the generated filename directly as the filename attribute.
            Default is False.

        Returns
        -------
        str
            The generated long RINEX filename.
        """

        # +++++ set extension
        if ext == "auto" and self.hatanka_input:
            ext = "crx"
        elif ext == "auto" and not self.hatanka_input:
            ext = "rnx"
        else:
            ext = "rnx"

        ext = "." + ext

        # +++++ set compression
        if compression == "auto" and self.compression:
            compression = "." + self.compression
        elif compression == "auto" and not self.compression:
            compression = ""
        elif (
            compression != ""
        ):  # when a manual compression arg is given, and is not void
            compression = "." + compression
        else:
            compression = ""

        # +++++ set file period and session
        self.mod_file_period(filename_style=filename_style)
        file_period_name = self.file_period
        session_name = self.session

        # +++++ set time format
        # default time format
        timeformat = "%Y%j%H%M"

        if file_period_name == "01D":  ## Daily case
            if session_name:
                timeformat = "%Y%j%H%M"
            else:
                timeformat = "%Y%j0000"  # Start of the day
        elif file_period_name[-1] == "M":  ### Subhourly case
            timeformat = "%Y%j%H%M"
        elif (
            file_period_name == "00U"
        ):  ## Unknown case: the filename deserves a full description to identify potential bug
            timeformat = "%Y%j%H%M"
        else:  ## Hourly case
            if filename_style in ("basic", "flex"):
                timeformat = "%Y%j%H00"  # Start of the hour
            elif filename_style == "exact":
                timeformat = "%Y%j%H%M"  # start of the minute

        longname = "_".join(
            (
                self.get_site(False, False),
                data_source,
                self.start_date.strftime(timeformat),
                file_period_name,
                self.sample_rate_string,
                self.get_sat_system() + obs_type + ext + compression,
            )
        )

        if inplace_set:
            self.filename = longname
            self.name_conv = "LONG"

        return longname

    def get_shortname(
        self,
        file_type="auto",
        compression="auto",
        filename_style="basic",
        inplace_set=False,
    ):
        """
        Generate the short RINEX filename.
        Can be stored directly as filename attribute with inplace=True.

        Parameters
        ----------
        file_type : str, optional
            'auto' (based on the RinexFile attribute) or manual: 'o', 'd', etc. Default is 'auto'.
        compression : str, optional
            'auto' (based on the RinexFile attribute) or manual: 'Z', 'gz', etc. Default is 'auto'.
            The value is given without a dot as the first character.
        filename_style : str, optional
            The style of the filename.
             Possible values are 'basic', 'flex', 'exact'.
             See the rinexmod main function's docstring/help for more details.
             Default is 'basic'.
        inplace_set : bool, optional
            If True, stores the generated filename directly as the filename attribute. Default is False.

        Returns
        -------
        str
            The generated short RINEX filename.
        """

        if file_type == "auto" and self.hatanka_input:
            file_type = "d"
        elif file_type == "auto" and not self.hatanka_input:
            file_type = "o"

        if compression == "auto":
            compression = self.compression

        if compression != "":
            compression = "." + compression

        self.mod_file_period(filename_style=filename_style)
        file_period_name = self.file_period
        session_name = self.session

        alphabet = list(map(chr, range(97, 123)))
        if file_period_name[-1] == "H":
            timeformat = (
                "%j" + alphabet[self.start_date.hour] + ".%y" + file_type + compression
            )
            start_date_use = self.start_date

        elif file_period_name[-1] == "M":
            timeformat = (
                "%j"
                + alphabet[self.start_date.hour]
                + "%M"
                + ".%y"
                + file_type
                + compression
            )
            start_date_use = round_time(
                self.start_date, timedelta(minutes=5), to="down"
            )

        else:  # regular case file_period_name == "01D"
            timeformat = "%j0.%y" + file_type + compression
            start_date_use = self.start_date

        shortname = self.get_site(True, True) + start_date_use.strftime(timeformat)

        if inplace_set:
            self.filename = shortname
            self.name_conv = "SHORT"

        return shortname

    def get_header_body(self, return_index_end=False):
        """
        Get the RINEX's header and body as a list of lines,
        and the index of the END OF HEADER line if asked
        """
        i_end = None
        i_end = search_idx_value(self.rinex_data, "END OF HEADER")

        head = self.rinex_data[: i_end + 1]
        body = self.rinex_data[i_end + 1 :]

        if return_index_end:
            return head, body, i_end
        else:
            return head, body

    # *****************************************************************************
    ### internal methods

    def _load_rinexdata(self, force_rnx_load=False):
        """
        Load the uncompressed rinex data into a list var using hatanaka library.
        Will return a table of lines of the uncompressed file, a 'name_conv' var
        that indicates if the file is named with the SHORT NAME convention or the
        LONG NAME convetion, and a status. Status 0 is OK. The other ones
        corresponds to the errors codes raised by rinexarchive and by rinexmod.
        01 - The specified file does not exists
        02 - Not an observation Rinex file
        03 - Invalid  or empty Zip file
        04 - Invalid Compressed Rinex file
        """

        ##### IF input is a file, checking if file exists
        if self.source_from_file and not os.path.isfile(self.path):
            rinex_data = None
            status = "01 - The specified file does not exists"
        ##### The input is a file
        else:
            try:
                rinex_data = hatanaka.decompress(self.content_input).decode("utf-8")
                rinex_data = rinex_data.split("\n")
                status = None

            except ValueError:
                rinex_data = None
                status = "03 - Invalid or empty compressed file"

            except hatanaka.hatanaka.HatanakaException:
                rinex_data = None
                status = "04 - Invalid Compressed RINEX file"

            if status:
                logger.error(status)

            if rinex_data:
                bool_l1_obs = "OBSERVATION DATA" in rinex_data[0]
                bool_l1_crx = "COMPACT RINEX FORMAT" in rinex_data[0]
            else:
                bool_l1_obs = False
                bool_l1_crx = False

            if not force_rnx_load and not (bool_l1_obs or bool_l1_crx):
                logger.error(
                    "File's 1st line does not match an Observation RINEX: "
                    + os.path.join(self.path)
                )
                logger.error("try to force the loading with force_rnx_load = True")
                rinex_data = None
                status = "02 - Not an observation RINEX file"

        if status:
            logger.error(status)

        return rinex_data, status

    #   _____      _     __  __      _   _               _
    #  / ____|    | |   |  \/  |    | | | |             | |
    # | |  __  ___| |_  | \  / | ___| |_| |__   ___   __| |___
    # | | |_ |/ _ \ __| | |\/| |/ _ \ __| '_ \ / _ \ / _` / __|
    # | |__| |  __/ |_  | |  | |  __/ |_| | | | (_) | (_| \__ \
    #  \_____|\___|\__| |_|  |_|\___|\__|_| |_|\___/ \__,_|___/
    #

    def get_naming_convention(self):
        """
        Get the naming convention based on a regular expression test
        """
        # Daily or hourly, hatanaka or not, gz or Z compressed file
        pattern_dic = regex_pattern_rinex_filename()

        pattern_shortname = re.compile(pattern_dic["shortname"])
        pattern_longname = re.compile(pattern_dic["longname"])
        # GFZ's DataCenter internal naming convention (here it is equivalent to a longname)
        pattern_longname_gfz = re.compile(pattern_dic["longname_gfz"])

        if pattern_shortname.match(os.path.basename(self.path)):
            name_conv = "SHORT"
        elif pattern_longname.match(os.path.basename(self.path)):
            name_conv = "LONG"
        elif pattern_longname_gfz.match(os.path.basename(self.path)):
            name_conv = "LONGGFZ"
        else:
            name_conv = "UNKNOWN"

        return name_conv

    def get_size(self):
        """
        Get RINEX file size
        """

        if self.status:
            return 0

        if self.source_from_file:
            size = os.path.getsize(self.path)
        else:
            size = 0

        return size

    def get_compression(self):
        """
        get the compression type in a 2-tuple: (compress,hatanaka)
        compress is None or a string: gz, Z, 7z
        hatanaka is a bool
        """

        basename = os.path.basename(self.path)

        # find compress value
        if basename.lower().endswith("z"):
            compress = os.path.splitext(basename)[1][1:]
        else:
            compress = None

        # find hatanaka value
        if self.name_conv == "SHORT":

            if compress:
                type_letter = basename.split(".")[-2][-1]
            else:
                type_letter = basename[-1]

            if type_letter == "d":
                hatanaka_bool = True
            else:
                hatanaka_bool = False

        else:  # LONG name
            if compress:
                type_ext = basename.split(".")[-2][-3:]
            else:
                type_ext = basename[-3:]

            if type_ext == "crx":
                hatanaka_bool = True
            else:
                hatanaka_bool = False

        return compress, hatanaka_bool

    def get_filename(self):
        """
        Get filename WITHOUT its compression extension
        """

        if self.status:
            return None

        if not self.compression:
            filename = os.path.basename(self.path)
        else:
            basename = os.path.splitext(os.path.basename(self.path))
            filename = basename[0]

        return filename

    def get_version(self):
        """
        Get RINEX version
        """

        if self.status:
            return None, None

        version_header_idx = search_idx_value(self.rinex_data, "RINEX VERSION / TYPE")
        version_header = self.rinex_data[version_header_idx]
        # Parse line
        rinex_ver_head_str = str(version_header[0:9].strip())
        rinex_ver_head_flt = float(
            re.search(r"[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)", rinex_ver_head_str)[0]
        )
        # must fit a regex pattern of a float, because nasty hidder characters can be present

        return rinex_ver_head_str, rinex_ver_head_flt

    def get_data_source(self):
        """
        Get data source: R, S, U from LONG filename, R per default
        """

        if self.status:
            return None

        if not "LONG" in self.name_conv:
            src = "R"
        else:
            src = self.filename[10]

        return src

    def _get_date_patterns(self):
        """
        Internal function to get the correct epoch pattern depending
        on the RINEX version

        Returns
        -------
        date_pattern : str
            a regex matching the epoch pattern.
        year_prefix : str
            for RINEX2, the year prefix.
        """
        # Date lines pattern
        if self.version_float >= 3.0:
            # Pattern of an observation line containing a date - RINEX 3
            # date_pattern = re.compile('> (\d{4}) (\d{2}) (\d{2}) (\d{2}) (\d{2}) ((?: |\d)\d.\d{4})')
            date_pattern = re.compile(
                r"> (\d{4}) (\d{2}| \d) (\d{2}| \d) (\d{2}| \d) (\d{2}| \d) ((?: |\d)\d.\d{4})"
            )
            year_prefix = ""  # Prefix of year for date formatting

        elif self.version_float < 3:
            # Pattern of an observation line containing a date - RINEX 2
            date_pattern = re.compile(
                r" (\d{2}) ((?: |\d)\d{1}) ((?: |\d)\d{1}) ((?: |\d)\d{1}) ((?: |\d)\d{1}) ((?: |\d)\d{1}.\d{4})"
            )
            year_prefix = "20"  # Prefix of year for date formatting
            ### !!!!!!!!! before 2000 must be implemented !!!!!!
        else:
            logger.warning("unable to find right RINEX version")
            date_pattern = None
            year_prefix = None

        return date_pattern, year_prefix

    def get_dates(self):
        """
        Get start and end date from rinex file.
        we search for the date of the first and last observation directly in
        the data.
        if you want the values in the header, use get_dates_in_header
        """

        if self.status:
            return None, None

        def _find_first_last_epoch_line(rnxobj, last_epoch=False):

            date_pattern, year_pattern = rnxobj._get_date_patterns()

            if last_epoch:
                datause = reversed(rnxobj.rinex_data)
            else:
                datause = rnxobj.rinex_data

            # Searching the last one of the file
            m = None
            for line in datause:
                m = re.search(date_pattern, line)
                if m:
                    break
            if m:
                year = year_pattern + m.group(1)

                # Building a date string
                epoc = (
                    year
                    + " "
                    + m.group(2)
                    + " "
                    + m.group(3)
                    + " "
                    + m.group(4)
                    + " "
                    + m.group(5)
                    + " "
                    + m.group(6)
                )

                epoc = datetime.strptime(epoc, "%Y %m %d %H %M %S.%f")
            else:
                epoc = None

            return epoc

        start_epoch = _find_first_last_epoch_line(self)
        end_epoch = _find_first_last_epoch_line(self, last_epoch=True)

        return start_epoch, end_epoch

    def get_dates_in_header(self):
        """
        Get start and end date from rinex file.
        Start date cames from TIME OF FIRST OBS file's header.
        In RINEX3, there's a TIME OF LAST OBS in the header but it's not available
        in RINEX2, so we search for the date of the last observation directly in
        the data.
        """

        if self.status:
            return None, None

        def _find_meta_label(meta_label_in):
            line = None
            line_found = None
            for line in self.rinex_data:
                if re.search(meta_label_in, line):
                    line_found = line
                    break
            # If not found
            if not line_found:
                date_out = None
            else:
                date_out = line.split()
                date_out = datetime.strptime(
                    " ".join(date_out[0:6]), "%Y %m %d %H %M %S.%f0"
                )
            return date_out

        start_head = _find_meta_label("TIME OF FIRST OBS")
        end_head = _find_meta_label("TIME OF LAST OBS")

        return start_head, end_head

    def get_dates_all(self):
        """
        Returns all the epochs in the RINEX
        """
        if self.status:
            return None, None

        samples_stack = []

        date_pattern, _ = self._get_date_patterns()

        for line in self.rinex_data:  # We get all the epochs dates
            if re.search(date_pattern, line):
                samples_stack.append(re.search(date_pattern, line))

        return samples_stack

    def get_sample_rate(self, plot=False):
        """
        Get sample rate from rinex file.
        The method returns 2 outputs: a str sample rate for the RINEX filenames, and the float value
        We get all the samples dates and get intervals. We then remove 0 ones (due to potential double samples).
        Then, we set the most frequent value as sample rate, and if more than 10% of the intervals are different
        from this sample rate, we set the sample rate as unknown (XXU).
        If there is not enought samples to compute an interval (less than two samples), we raise an error
        with error code 5. If there is only two samples, i.e. one interval, we set the sample rate to
        unknown because we can not compute a most frequent interval.
        We then round the obtained value and translate it to a rinex 3 longname compliant format.
        If plot is set to True, will plot the samples intervals.
        """

        if self.status:
            return None, None

        samples_stack = self.get_dates_all()
        date_pattern, year_prefix = self._get_date_patterns()

        # If less than 2 samples, can't get a sample rate
        if len(samples_stack) < 2:
            self.status = "05 - Less than two epochs in the file"
            logger.error(
                "less than 2 samples found, can't get a sample rate %s",
                samples_stack,
            )
            return None, None

        # Building a date string
        def _date_conv(sample):
            date = (
                year_prefix
                + sample.group(1)
                + " "
                + sample.group(2)
                + " "
                + sample.group(3)
                + " "
                + sample.group(4)
                + " "
                + sample.group(5)
                + " "
                + sample.group(6)
            )

            date = datetime.strptime(date, "%Y %m %d %H %M %S.%f")
            return date

        # Format dates to datetime
        samples_stack = [_date_conv(d) for d in samples_stack]
        samples_rate_diff = np.diff(np.array(samples_stack))  # Getting intervals
        # Converting timedelta to seconds and removing 0 values (potential doubles in epochs)
        samples_rate_diff = [
            diff.total_seconds()
            for diff in samples_rate_diff
            if diff != timedelta(seconds=0)
        ]

        # If less than one interval after removing 0 values, can't get a sample rate
        if len(samples_rate_diff) < 1:
            self.status = "05 - Less than two epochs in the file"
            logger.error(
                "less than one interval after removing 0 values %s",
                samples_rate_diff,
            )
            return None, None

        # If less than 2 intervals, can't compare intervals
        if len(samples_rate_diff) < 2:
            return "00U", 0.0

        # Most frequent
        sample_rate_num = max(set(samples_rate_diff), key=samples_rate_diff.count)

        # Counting the intervals that are not equal to the most frequent
        num_bad_sp = len(
            [diff for diff in samples_rate_diff if diff != sample_rate_num]
        )

        non_nominal_interval_percent = num_bad_sp / len(samples_rate_diff)

        plot = False
        if plot:
            import matplotlib.pyplot as plt

            print(
                "{:29} : {}".format(
                    "Sample intervals not nominals",
                    str(non_nominal_interval_percent * 100) + " %",
                )
            )
            plt.plot(samples_rate_diff)
            plt.show()

        non_nominal_trigger = 0.45
        if (
            non_nominal_interval_percent > non_nominal_trigger
        ):  # Don't set sample rate to files
            # That have more that 45% of non nominal sample rate
            logger.error(
                "non nominal sample rate >%d%%: %d%% (# epochs: %d)",
                non_nominal_trigger * 100,
                non_nominal_interval_percent * 100,
                len(samples_stack),
            )
            return "00U", 0.0

        # Format of sample rate from RINEX3 specs : RINEX Long Filenames
        # We round samples rates to avoid leap-seconds related problems
        if sample_rate_num <= 0.0001:
            # XXU – Unspecified
            sample_rate_str = "00U"
        elif sample_rate_num <= 0.01:
            # XXC – 100 Hertz
            sample_rate_num = round(sample_rate_num, 4)
            sample_rate_str = (str(int(1 / (100 * sample_rate_num))) + "C").rjust(
                3, "0"
            )
        elif sample_rate_num < 1:
            # XXZ – Hertz
            sample_rate_num = round(sample_rate_num, 2)
            sample_rate_str = (str(int(1 / sample_rate_num)) + "Z").rjust(3, "0")
        elif sample_rate_num < 60:
            # XXS – Seconds
            sample_rate_num = round(sample_rate_num, 0)
            sample_rate_str = (str(int(sample_rate_num)) + "S").rjust(3, "0")
        ##### NB: a sample rate at the minute level or even above is very unlikely
        elif sample_rate_num < 3600:
            # XXM – Minutes
            sample_rate_num = round(sample_rate_num, 0)
            sample_rate_str = (str(int(sample_rate_num / 60)) + "M").rjust(3, "0")
        elif sample_rate_num < 86400:
            # XXH – Hours
            sample_rate_num = round(sample_rate_num, 0)
            sample_rate_str = (str(int(sample_rate_num / 3600)) + "H").rjust(3, "0")
        elif sample_rate_num <= 8553600:
            # XXD – Days
            sample_rate_num = round(sample_rate_num, 0)
            sample_rate_str = (str(int(sample_rate_num / 86400)) + "D").rjust(3, "0")
        else:
            # XXU – Unspecified
            sample_rate_str = "00U"

        return sample_rate_str, sample_rate_num

    def get_file_period_from_filename(self):
        """
        Get the file period from the file's name.
        In long name convention, gets it striaght from the file name.
        In short name convention, traduces digit to '01H' and '0' to 01D
        """

        if self.status:
            return None, None

        session = False

        if self.name_conv == "SHORT":
            file_period = self.filename[7:8]
            if file_period.isdigit():
                if file_period != "0":
                    session = True
                # 01D–1 Day
                file_period = "01D"
            elif file_period.isalpha():
                # 01H–1 Hour
                file_period = "01H"
            else:
                # 00U-Unspecified
                file_period = "00U"

        elif self.name_conv == "LONG":
            file_period = self.filename[24:27]

        elif self.name_conv == "LONGGFZ":
            file_period = self.filename[46:49]
        else:
            # 00U-Unspecified
            file_period = "00U"

        return file_period, session

    #         the RINEX file period is tolerant and stick to the actual data content,
    #         but then can be odd (e.g. 07H, 14H...). A strict file period is applied
    #         per default (01H or 01D), being compatible with the IGS conventions

    def get_file_period_from_data(self, inplace_set=False):
        """
        Get the file period from the data themselves.

        see also mod_file_period_basic()
        to round this value to a conventional one

        Parameters
        ----------
        inplace_set : bool, optional
            change the values in the RinexFile object. The default is False.

        Returns
        -------
        file_period : str
            the file period: '01H', '01D', '00U'...
        session : bool
            RINEX is a hourly session or not.

        """

        file_period, session = file_period_from_timedelta(
            self.start_date, self.end_date
        )

        if inplace_set:
            self.file_period = file_period
            self.session = session

        return file_period, session

    def mod_file_period(self, filename_style="basic"):

        self.get_file_period_from_data(inplace_set=True)

        if filename_style == "basic":
            self.mod_file_period_basic()
        elif filename_style in ("flex", "exact"):
            pass  ## it is the same as get_file_period_from_data()
        else:
            logger.error(
                "style %s not recognized. Accepts only 'basic', 'flex', 'exact'",
                filename_style,
            )

        return self.file_period, self.session

    def mod_file_period_basic(self):
        """
        Round the RINEX file period to a basic conventional value: 01D, 01H

        NB: this method aims to respect the IGS convention and thus uses NOMINAL
        period

        For the time beeing, it does not handle the subhourly case

        Returns
        -------
        file_period_rnd : str
            the file period: '01H', '01D', '00U'...
        session_rnd : bool or None
            the RINEX is an hourly session or not.

        """

        if self.file_period[2] == "H" and int(self.file_period[:2]) > 1:
            file_period_rnd = "01D"
            session_rnd = False
        else:
            file_period_rnd = self.file_period
            session_rnd = self.session

        self.file_period = file_period_rnd
        self.session = session_rnd

        return file_period_rnd, session_rnd

    def get_sat_system(self):
        """
        Parse RINEX VERSION / TYPE line to get observable type
        """

        if self.status:
            return None

        # Identify line that contains RINEX VERSION / TYPE
        sat_system_header_idx = search_idx_value(
            self.rinex_data, "RINEX VERSION / TYPE"
        )
        sat_system_head = self.rinex_data[sat_system_header_idx]
        # Parse line
        sat_system = sat_system_head[40:41]

        return sat_system

    def get_site_from_filename(self, lower_case=True, only_4char=False):
        """
        Get site name from the filename
        """

        if self.status:
            return None

        if self.name_conv in ("LONG", "LONGGFZ"):
            site_out = self.filename[:9]
        else:
            site_out = self.filename[:4] + "00" + "XXX"

        if only_4char:
            site_out = site_out[:9]

        if lower_case:
            return site_out.lower()
        else:
            return site_out.upper()

    def get_site_from_header(self):
        """
        Get site name from the MARKER NAME line in rinex file's header
        """

        if self.status:
            return ""

        site_head = "MARKER NAME"

        for line in self.rinex_data:
            if re.search(site_head, line):
                site_head = line
                break

        if site_head == "MARKER NAME":
            return None

        site_head = site_head.split(" ")[0].upper()

        return site_head

    def get_sys_obs_types(self):
        """
        Get the systems/observables values from the
        ``SYS / # / OBS TYPES`` lines

        for RINEX 3/4 only

        Returns
        -------
        dict_sys_obs : dict
            a dictionary of lists describing the observables per system, e.g.:
            ``{"G" : ["C1C","C1W","C2W","L1C","L2W","S1C",S2W"]}``.

        dict_sys_nobs : dict
            a dictionary of integer describing the number of observables per system, e.g.:
            ``{'C': 8, 'E': 16, 'G': 16, 'I': 4, 'R': 12, 'S': 4}``.

        """

        if self.status:
            return

        if self.version_float < 3:
            logger.warn("get_sys_obs_types is only compatible with RINEX3/4")
            return

        # Identify line that contains SYS / # / OBS TYPES
        sys_obs_idx0 = search_idx_value(self.rinex_data, "SYS / # / OBS TYPES")

        sys_obs_idx_fin = sys_obs_idx0
        while "SYS / # / OBS TYPES" in self.rinex_data[sys_obs_idx_fin]:
            sys_obs_idx_fin += 1

        #### get the systems and observations
        lines_sys = self.rinex_data[sys_obs_idx0:sys_obs_idx_fin]

        ## clean SYS / # / OBS TYPES
        lines_sys = [l[:60] for l in lines_sys]

        ## manage the 2 lines systems => they are stacked in one
        for il, l in enumerate(lines_sys):
            if l[0] == " ":
                lines_sys[il - 1] = lines_sys[il - 1] + l
                lines_sys.remove(l)

        #### store system and observables in a dictionnary
        dict_sys_obs = dict()
        dict_sys_nobs = dict()

        for il, l in enumerate(lines_sys):
            sysobs = l.split()
            sys = sysobs[0]
            dict_sys_obs[sys] = sysobs[2:]
            dict_sys_nobs[sys] = int(sysobs[1])
            ## adds the LLI and SSI indicators
            if len(sysobs[2:]) != int(sysobs[1]):
                logger.warn(
                    "difference between theorectical (%d) and actual (%d) obs nbr for sys (%s)",
                    len(sysobs[2:]),
                    int(sysobs[1]),
                    sys,
                )

        return dict_sys_obs, dict_sys_nobs

    #  __  __           _   __  __      _   _               _
    # |  \/  |         | | |  \/  |    | | | |             | |
    # | \  / | ___   __| | | \  / | ___| |_| |__   ___   __| |___
    # | |\/| |/ _ \ / _` | | |\/| |/ _ \ __| '_ \ / _ \ / _` / __|
    # | |  | | (_) | (_| | | |  | |  __/ |_| | | | (_) | (_| \__ \
    # |_|  |_|\___/ \__,_| |_|  |_|\___|\__|_| |_|\___/ \__,_|___/

    ### ***************************************************************************
    ### mod methods. change the content of the RINEX header

    def mod_marker(self, marker_inp=None, number_inp=None):
        """
        Modify within the RINEX header the marker
        (``MARKER NAME`` line)

        Parameters
        ----------
        marker_inp : str, optional
            site marker ID a.k.a. station code (4 or 9 characters).
            The default is None.
        number_inp : str, optional
            DOMES number. The default is None.

        Returns
        -------
        None.
        """

        if self.status:
            return

        ###marker_inp is a mandatory arguement, no None!
        if not marker_inp and not number_inp:
            return

        marker_name_header_idx = search_idx_value(self.rinex_data, "MARKER NAME")
        # Identify line that contains MARKER NAME
        if marker_inp:
            # Edit line
            new_line = "{}".format(marker_inp.ljust(60)) + "MARKER NAME"
            if marker_name_header_idx:
                # marker_name_meta = self.rinex_data[marker_name_header_idx]
                # Set line
                self.rinex_data[marker_name_header_idx] = new_line
            else:
                pgm_header_idx = search_idx_value(
                    self.rinex_data, "PGM / RUN BY / DATE"
                )
                self.rinex_data.insert(pgm_header_idx, new_line)

        if number_inp:
            # Identify line that contains MARKER NUMBER
            ## marker_number_header_idx = next((i for i, e in enumerate(self.rinex_data) if 'MARKER NUMBER' in e), None)
            marker_number_header_idx = search_idx_value(
                self.rinex_data, "MARKER NUMBER"
            )
            # Edit line
            new_line = "{}".format(number_inp.ljust(60)) + "MARKER NUMBER"
            if marker_number_header_idx:  # The line exsits
                # Set line
                self.rinex_data[marker_number_header_idx] = new_line
            else:  # The line does not exsits
                # Set line
                self.rinex_data.insert(marker_name_header_idx + 1, new_line)

        return

    def mod_receiver(self, serial=None, type=None, firmware=None, keep_rnx_rec=False):
        """
        Modify within the RINEX header the receiver information
        (``REC # / TYPE / VERS`` line)

        Parameters
        ----------
        serial : str, optional
            Receiver Serial Number. The default is None.
        type : str, optional
            Receiver model. The default is None.
        firmware : str, optional
            Firmware version. The default is None.
        keep_rnx_rec : bool, optional
            Keep the RINEX receiver header record in the output RINEX.
            Metadata from the external source (e.g. sitelogs) will not be modded.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if not any([serial, type, firmware]):
            return

        # Identify line that contains REC # / TYPE / VERS
        receiver_header_idx = search_idx_value(self.rinex_data, "REC # / TYPE / VERS")
        receiver_head = self.rinex_data[receiver_header_idx]
        # Parse line
        serial_head = receiver_head[0:20]
        type_head = receiver_head[20:40]
        firmware_head = receiver_head[40:60]
        label_head = receiver_head[60:]

        # warning: for the receiver, info in te input RINEX might be the correct ones
        def _mod_rec_check(field_type, rinex_val, metadata_val):
            if metadata_val and rinex_val.strip() != metadata_val.strip():
                logger.warning(
                    "%s rec. %s in RINEX (%s) & in metadata (%s) are different.",
                    self.get_site(lower_case=False),
                    field_type,
                    rinex_val.strip(),
                    metadata_val.strip(),
                )
                logger.warning(
                    "The RINEX value might be the correct one, double-check your metadata source."
                )
                return True
                # True if the check fails: counter-intuitive, but makes sense for the test below
            else:
                return False

        rec_chk_sn = _mod_rec_check(
            "serial number", rinex_val=serial_head, metadata_val=serial
        )
        rec_chk_mt = _mod_rec_check(
            "model type", rinex_val=type_head, metadata_val=type
        )
        rec_chk_fw = _mod_rec_check(
            "firmware version", rinex_val=firmware_head, metadata_val=firmware
        )

        if keep_rnx_rec and (rec_chk_sn or rec_chk_fw or rec_chk_fw):
            logger.info(
                "RINEX & metadata are different, but receiver values are kept (keep_rnx_rec = True)"
            )
            return

        # Edit line
        if serial:
            serial_head = str(serial)[:20].ljust(20)
        if type:
            type_head = str(type)[:20].ljust(20)
        if firmware:
            firmware_head = str(firmware)[:20].ljust(20)
        new_line = serial_head + type_head + firmware_head + label_head
        # Set line
        self.rinex_data[receiver_header_idx] = new_line

        return

    def mod_antenna(self, serial=None, type=None):
        """
        Modify within the RINEX header the antenna information
        (``ANT # / TYPE`` line)

        Parameters
        ----------
        serial : str, optional
            Antenne Serial Number. The default is None.
        type : str, optional
            Antenna model. The default is None.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if not any([serial, type]):
            return

        # Identify line that contains ANT # / TYPE
        antenna_header_idx = search_idx_value(self.rinex_data, "ANT # / TYPE")
        antenna_head = self.rinex_data[antenna_header_idx]
        # Parse line
        serial_head = antenna_head[0:20]
        type_head = antenna_head[20:40]
        label_head = antenna_head[60:]
        # Edit line
        if serial:
            serial_head = str(serial)[:20].ljust(20)
        if type:
            type_head = str(type)[:20].ljust(20)
        new_line = serial_head + type_head + " " * 20 + label_head
        # Set line
        self.rinex_data[antenna_header_idx] = new_line

        return

    def mod_interval(self, sample_rate_input=None):
        """
        Modify within the RINEX header the data sample rate
        (``INTERVAL`` line)

        Parameters
        ----------
        sample_rate_input : TYPE, optional
            Data sample rate. The default is None.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if not any([sample_rate_input]):
            return

        # Identify line that contains INTERVAL
        line_exists = False
        idx = -1
        interval_idx = None

        for e in self.rinex_data:
            idx += 1
            if "INTERVAL" in e:
                line_exists = True
                interval_idx = idx
                break
            elif "TIME OF FIRST OBS" in e:
                interval_idx = idx
            elif "END OF HEADER" in e:
                break
            else:
                continue

        if line_exists:
            # interval_idx = next(i for i, e in enumerate(self.rinex_data) if 'INTERVAL' in e)
            interval_head = self.rinex_data[interval_idx]
            label_head = interval_head[60:]
        else:
            # interval_idx = next(i for i, e in enumerate(self.rinex_data) if 'TIME OF FIRST OBS' in e)
            label_head = "INTERVAL"

        # Parse line
        sample_rate_head = "{:10.3f}".format(float(sample_rate_input))

        new_line = sample_rate_head + " " * 50 + label_head

        # Set line
        if line_exists:
            self.rinex_data[interval_idx] = new_line
        else:
            self.rinex_data.insert(interval_idx, new_line)

        return

    def mod_antenna_pos(self, X=None, Y=None, Z=None):
        """
        Modify within the RINEX header the X Y Z RINEX's approximative position.
        (``APPROX POSITION XYZ`` line)

        Parameters
        ----------
        X,Y,Z : float, optional
            X Y Z position. The default is None.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if (X is None) and (Y is None) and (X is None):
            return

        # Identify line that contains APPROX POSITION XYZ
        antenna_pos_header_idx = search_idx_value(
            self.rinex_data, "APPROX POSITION XYZ"
        )
        antenna_pos_head = self.rinex_data[antenna_pos_header_idx]
        # Parse line
        x_head = antenna_pos_head[0:14]
        y_head = antenna_pos_head[14:28]
        z_head = antenna_pos_head[28:42]
        label = antenna_pos_head[60:]
        # Edit line
        if (
            X is not None
        ):  # Format as 14.4 float. Set to zero if too large but should not happen
            x_head = "{:14.4f}".format(float(X))
            if len(x_head) > 14:
                x_head = "{:14.4f}".format(float("0"))
        if Y is not None:
            y_head = "{:14.4f}".format(float(Y))
            if len(y_head) > 14:
                y_head = "{:14.4f}".format(float("0"))
        if Z is not None:
            z_head = "{:14.4f}".format(float(Z))
            if len(z_head) > 14:
                z_head = "{:14.4f}".format(float("0"))
        new_line = x_head + y_head + z_head + " " * 18 + label
        # Set line
        self.rinex_data[antenna_pos_header_idx] = new_line

        return

    def mod_antenna_delta(self, H=None, E=None, N=None):
        """
        Modify within the RINEX header the H E N antenna's excentricity
        (``ANTENNA: DELTA H/E/N`` line).

        Parameters
        ----------
        H, E, N: float, optional
            H E N position. The default is None.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if (H is None) and (E is None) and (N is None):
            return

        # Identify line that contains ANTENNA: DELTA H/E/N
        antenna_delta_header_idx = search_idx_value(
            self.rinex_data, "ANTENNA: DELTA H/E/N"
        )
        antenna_delta_head = self.rinex_data[antenna_delta_header_idx]
        # Parse line
        h_head = antenna_delta_head[0:14]
        e_head = antenna_delta_head[14:28]
        n_head = antenna_delta_head[28:42]
        label = antenna_delta_head[60:]
        # Edit line
        if (
            H is not None
        ):  # Format as 14.4 float. Set to zero if too large but should not happen
            h_head = "{:14.4f}".format(float(H))
            if len(h_head) > 14:
                h_head = "{:14.4f}".format(float("0"))
        if E is not None:
            e_head = "{:14.4f}".format(float(E))
            if len(e_head) > 14:
                e_head = "{:14.4f}".format(float("0"))
        if N is not None:
            n_head = "{:14.4f}".format(float(N))
            if len(n_head) > 14:
                n_head = "{:14.4f}".format(float("0"))
        new_line = h_head + e_head + n_head + " " * 18 + label
        # Set line
        self.rinex_data[antenna_delta_header_idx] = new_line

        return

    def mod_agencies(self, operator=None, agency=None):
        """
        Modify within the RINEX header the ``OBSERVER / AGENCY`` line

        Parameters
        ----------
        operator : str, optional
            Operator. The default is None.
        agency : str, optional
            Agency. The default is None.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if not any([operator, agency]):
            return

        # Identify line that contains OBSERVER / AGENCY
        agencies_header_idx = search_idx_value(self.rinex_data, "OBSERVER / AGENCY")

        if not agencies_header_idx:  ##### THIS TEST SHOULD BE IN ALL MOD METHODS !!!!
            logger.warning(
                "no %s field has been found in %s, unable to mod it",
                "OBSERVER / AGENCY",
                self.filename,
            )
            return

        agencies_head = self.rinex_data[agencies_header_idx]
        # Parse line
        operator_head = agencies_head[0:20]
        agency_head = agencies_head[20:40]
        label_head = agencies_head[60:]
        # Edit line
        if operator:  # Format as 14.4 float. Cut if too large but will not happen
            operator_head = operator[:20].ljust(20)
        if agency:
            agency_head = agency[:40].ljust(40)
        new_line = operator_head + agency_head + label_head
        # Set line
        self.rinex_data[agencies_header_idx] = new_line

        return

    def mod_sat_system(self, sat_system):
        """
        Modify within the RINEX header the *type* of the
        ``RINEX VERSION / TYPE`` line


        Parameters
        ----------
        sat_system :
            Satellite system.

        Returns
        -------
        None.

        """

        if self.status:
            return

        if not sat_system:
            return

        if "+" in sat_system:  ### case MIXED system
            sat_system = "MIXED"
            sat_system_code = "M"
        else:  ### case single system
            gnss_codes = {
                "GPS": "G",
                "GLO": "R",
                "GAL": "E",
                "BDS": "C",
                "QZSS": "J",
                "IRNSS": "I",
                "SBAS": "S",
                "MIXED": "M",
            }
            sat_system_code = gnss_codes.get(sat_system)

            if not sat_system_code:
                sat_system_code = sat_system
                sat_system = ""

        ### rewrite the RINEX header line

        # Identify line that contains RINEX VERSION / TYPE
        sat_system_header_idx = search_idx_value(
            self.rinex_data, "RINEX VERSION / TYPE"
        )
        sat_system_head = self.rinex_data[sat_system_header_idx]
        # Parse line
        rinex_ver_head = sat_system_head[0:9]
        type_of_rinex_file_head = sat_system_head[20:40]
        # sat_system_head = sat_system_head[40:60]
        label = sat_system_head[60:]

        sat_system_head = sat_system_code[0] + " : " + sat_system[:16].ljust(16)
        new_line = (
            rinex_ver_head
            + " " * 11
            + type_of_rinex_file_head
            + sat_system_head
            + label
        )
        # Set line
        self.rinex_data[sat_system_header_idx] = new_line

        return

    def mod_time_obs(self, first_obs=None, last_obs=None):
        """
        Modify within the RINEX header the ``TIME OF FIRST OBS``
        and ``TIME OF LAST OBS``


        Parameters
        ----------
        first_obs : datetime, optional
            epoch of first observation. The default is None.
        last_obs : datetime, optional
            epoch of last observation. The default is None.

        Returns
        -------
        None

        """

        if self.status:
            return

        if not first_obs:
            first_obs = self.start_date

        if not last_obs:
            last_obs = self.end_date

        # Identify line that contains TIME OF FIRST OBS
        first_obs_idx = search_idx_value(self.rinex_data, "TIME OF FIRST OBS")
        last_obs_idx = search_idx_value(self.rinex_data, "TIME OF LAST OBS")

        first_obs_head = self.rinex_data[first_obs_idx]

        # Parse line
        def _time_line_make(time, sys="GPS"):
            timel = "{:6d}{:6d}{:6d}{:6d}{:6d}{:13.7f}     {:3s}        "

            timelout = timel.format(
                time.year,
                time.month,
                time.day,
                time.hour,
                time.minute,
                time.second + time.microsecond * 10**-6,
                sys,
            )

            return timelout

        sysuse = first_obs_head[48:52]
        line_firstobs = _time_line_make(first_obs, sysuse) + "TIME OF FIRST OBS"
        line_lastobs = _time_line_make(last_obs, sysuse) + "TIME OF LAST OBS"

        self.rinex_data[first_obs_idx] = line_firstobs
        if last_obs_idx:
            self.rinex_data[last_obs_idx] = line_lastobs
        else:
            self.rinex_data.insert(first_obs_idx + 1, line_lastobs)

        return

    def mod_sys_obs_types(self, dict_sys_obs):
        """
        Modify within the RINEX header the systems/observables
        of the ``SYS / # / OBS TYPES`` lines

        for RINEX 3/4 only

        Parameters
        ----------
        dict_sys_obs : dict
            a dictionary of lists describing the observables per system, e.g.:
            ``{"G" : ["C1C","C1W","C2W","L1C","L2W","S1C",S2W"]}``
        """

        if self.status:
            return

        if not any([dict_sys_obs]):
            return

        if self.version_float < 3:
            logger.warn("mod_sys_obs_types is only compatible with RINEX3/4")
            return

        # Identify line that contains SYS / # / OBS TYPES
        sys_obs_idx0 = search_idx_value(self.rinex_data, "SYS / # / OBS TYPES")

        sys_obs_idx_fin = sys_obs_idx0
        while "SYS / # / OBS TYPES" in self.rinex_data[sys_obs_idx_fin]:
            sys_obs_idx_fin += 1

        ### parse the dict
        lfmt_stk = []
        for sys, obs in dict_sys_obs.items():
            ### make slice of 13 elements, i.e. the number of obervable max in
            ### one line
            obs_slice = slice_list(obs, 13)

            for iobss, obss in enumerate(obs_slice):
                nobss = len(obss)
                if iobss == 0:
                    l = "{:1s}  {:3d}" + " {:3s}" * nobss + "    " * (13 - nobss)
                    fmt_tup = tuple([sys, nobss] + obss)
                    lfmt = l.format(*fmt_tup)
                else:
                    l = "      " + " {:3s}" * nobss + "   " * (13 - nobss)
                    lfmt = l.format(*obss)
                lfmt = lfmt + "  SYS / # / OBS TYPES"
                lfmt_stk.append(lfmt)

        self.rinex_data = (
            self.rinex_data[:sys_obs_idx0]
            + lfmt_stk
            + self.rinex_data[sys_obs_idx_fin:]
        )

        return

    def mod_filename_data_freq(self, data_freq_inp=None):
        """
        Modify within the RINEX filename the data freqency
        (``30S``, ``01S``...)
        """

        if not data_freq_inp:
            return

        self.sample_rate_str = data_freq_inp
        return

    def mod_filename_file_period(self, file_period_inp=None):
        """
        Modify within the RINEX filename the period
        (``01D``, ``01H``, ``15M``...)
        """

        if not file_period_inp:
            return

        self.file_period = file_period_inp
        return

    def mod_filename_data_source(self, data_source_inp=None):
        """
        Modify within the RINEX filename the data source
        (``R``, ``S``, ``U``)
        """
        if not data_source_inp:
            return

        self.data_source = data_source_inp
        return

    #############################################################################
    ### misc methods. change the content of the RINEX header

    def get_as_string(self, encode="utf-8"):
        """
        get the RINEX data (header and body) as a string
        """
        return "\n".join(self.rinex_data).encode(encode)

    def write_to_path(self, output_directory, compression="gz", no_hatanaka=True):
        """
        Will turn rinex_data from list to string, utf-8, then compress as hatanaka
        and zip to the 'compression' format, then write to file. The 'compression' param
        will be used as an argument to hatanaka.compress and for naming the output file.
        Available compressions are those of hatanaka compress function :

        Parameters
        ----------
        output_directory : str
            The output directory.

        compression : str, optional
            'gz' (default), 'bz2', 'Z',
            'none' (string, compliant with hatanaka module) or
            None (NoneType, compliant with the rinex object initialisation).
            The default is 'gz'.

        no_hatanaka : bool, optional
            If True, the Hatanaka compression is not performed.
            (Hatanaka compression is applied per default.)
            The default is False.

        Returns
        -------
        outputfile : str
            The output RINEX file path.

        """

        if self.status:
            return

        # make the compression compliant with hatanaka module
        # (accept only 'none' as string)
        if not compression:
            comp_htnk_inp = "none"
        else:
            comp_htnk_inp = compression

        output_data = self.get_as_string()

        if not no_hatanaka:  ## regular case, Hatanaka-compression of the RINEX
            output_data = hatanaka.compress(output_data, compression=comp_htnk_inp)

        ### The data source is an actual RINEX file
        if self.source_from_file:
            if not no_hatanaka:
                # manage hatanaka compression extension
                # RNX3
                if "rnx" in self.filename:
                    filename_out = self.filename.replace("rnx", "crx")
                # RNX2
                elif self.filename[-1] in "o":
                    filename_out = self.filename[:-1] + "d"
                else:
                    filename_out = self.filename
            else:  ### NO Hatanaka-compression of the RINEX, the extension must be changed
                # RNX3
                if "crx" in self.filename:
                    filename_out = self.filename.replace("crx", "rnx")
                # RNX2
                elif self.filename[-1] in "d":
                    filename_out = self.filename[:-1] + "o"
                else:
                    filename_out = self.filename

            # manage low-level compression extension
            if compression in ("none", None):
                outputfile = os.path.join(output_directory, filename_out)
            else:
                outputfile = os.path.join(
                    output_directory, filename_out + "." + compression
                )
        ### the data source is a StringIO
        else:
            outputfile = output_directory

        Path(outputfile).write_bytes(output_data)

        self.path_output = outputfile

        return outputfile

    def add_comment(self, comment=None, add_as_first=False):
        """
        Add a comment line to the end of the RINEX header.

        Parameters
        ----------
        comment : str, optional
            The comment to be added. If None, the function returns immediately.
        add_as_first : bool, optional
            If True, the comment is added as the first comment. Default is False.

        Returns
        -------
        None

        Notes
        -----
        If no comments are in the header, they will be added just before the 'END OF HEADER'.
        Nevertheless, the sort_header() method executed a bit after will bring them
        just after PGM '/ RUN BY / DATE'
        """
        if self.status:
            return

        if not comment:
            return

        end_of_header_idx = search_idx_value(self.rinex_data, "END OF HEADER") + 1
        idx = [
            i
            for i, e in enumerate(self.rinex_data[0:end_of_header_idx])
            if "COMMENT" in e
        ]

        if (
            len(comment) < 60
        ):  # if the comment is shorter than 60 characters, we center it with dashes
            new_line = " {} ".format(comment).center(59, "-")[:59] + " COMMENT"
        else:  # if the comment is longer than 60 characters, we print it as it is (truncated to 60 characters)
            new_line = comment[:59] + " COMMENT"

        # regular case: some comments already exist
        if len(idx) > 0:
            # add the comment as last (default)
            if not add_as_first:
                last_comment_idx = max(idx)
                new_comment_idx = last_comment_idx + 1
            # add the comment as first
            else:
                first_comment_idx = min(idx)
                new_comment_idx = first_comment_idx
        # no comment in the header, then the first comment is added before the 'END OF HEADER'
        else:
            new_comment_idx = end_of_header_idx - 1

        self.rinex_data.insert(new_comment_idx, new_line)

        return

    def add_prg_run_date_comment(self, program, run_by):
        """
        Add a COMMENT, looking like as a 'PGM / RUN BY / DATE'-like line
        Useful to describe autorino edition, but without erasing the conversion
        program information
        """
        if self.status:
            return

        date = datetime.utcnow().strftime("%Y%m%d %H%M%S UTC")
        new_line = "{:20}{:20}{:20}{:}".format(program, run_by, date, "COMMENT")

        self.add_comment(new_line, add_as_first=True)

    def add_comments(self, comment_list):
        """
        Add several comments at the same time.
        The input is then a list of comments (strings)
        Useful for the full history changes for instance

        """
        for com in comment_list:
            self.add_comment(com)
        return

    def clean_history(self):
        """
        Remove the Full History Comments
        """

        i_start, i_end = None, None

        for il, l in enumerate(self.rinex_data):
            if "FULL HISTORY" in l:
                i_start = il
            if "Removed" in l:
                i_end = il
            if "END OF HEADER" in l:
                break

        if i_start and i_end:
            self.rinex_data = (
                self.rinex_data[0 : i_start - 1] + self.rinex_data[i_end + 1 :]
            )
        return

    def clean_rinexmod_comments(self, clean_history=True):
        """
        Remove all RinexMod Comments
        """
        if clean_history:
            self.clean_history()

        rinex_data_new = []
        for l in self.rinex_data:
            if not "rinexmod" in l.lower():
                rinex_data_new.append(l)
        self.rinex_data = rinex_data_new
        return

    def clean_translation_comments(
        self, internal_use_only=True, format_conversion=False
    ):
        """
        clean warning blocks generated during RINEX 2>3 translation

        Parameters
        ----------
        internal_use_only : bool, optional
            clean ``"WARNING - FOR INTERNAL USE ONLY"`` block.
            The default is True.
        format_conversion : bool, optional
            clean ``WARNING - FORMAT CONVERSION`` block.
            The default is False.

        Returns
        -------
        None.

        """

        def __translat_cleaner(rinex_data_in, block_title):
            i_start, i_end = None, None
            inblock = False
            for il, l in enumerate(rinex_data_in):
                if block_title in l:
                    i_start = il - 1  ## -1 for the 1st "*********" line
                    inblock = True
                if inblock and "*" * 50 in l:
                    i_end = il
                    inblock = False
                    break
                if "END OF HEADER" in l:
                    break

            if i_start and i_end:
                rinex_data_out = (
                    rinex_data_in[0 : i_start - 1] + rinex_data_in[i_end + 1 :]
                )
            else:
                logger.warn("no block %s found in header", block_title)
                rinex_data_out = rinex_data_in

            return rinex_data_out

        if internal_use_only:
            self.rinex_data = __translat_cleaner(
                self.rinex_data, "WARNING - FOR INTERNAL USE ONLY"
            )
        if format_conversion:
            self.rinex_data = __translat_cleaner(
                self.rinex_data, "WARNING - FORMAT CONVERSION"
            )

        return

    def sort_header(self):
        header_order = [
            "RINEX VERSION / TYPE",
            "PGM / RUN BY / DATE",
            "COMMENT",
            "MARKER NAME",
            "MARKER NUMBER",
            "MARKER TYPE",
            "OBSERVER / AGENCY",
            "REC # / TYPE / VERS",
            "ANT # / TYPE",
            "APPROX POSITION XYZ",
            "ANTENNA: DELTA H/E/N",
            "ANTENNA: DELTA X/Y/Z",
            "ANTENNA: PHASECENTER",
            "ANTENNA: B.SIGHT XYZ",
            "ANTENNA: ZERODIR AZI",
            "ANTENNA: ZERODIR XYZ",
            "CENTER OF MASS: XYZ",
            "DOI",
            "LICENSE OF USE",
            "STATION INFORMATION",
            "SYS / # / OBS TYPES",
            "SIGNAL STRENGTH UNIT",
            "INTERVAL",
            "TIME OF FIRST OBS",
            "TIME OF LAST OBS",
            "RCV CLOCK OFFS APPL",
            "SYS / DCBS APPLIED",
            "SYS / PCVS APPLIED",
            "SYS / SCALE FACTOR",
            "SYS / PHASE SHIFT",
            "GLONASS SLOT / FRQ #",
            "GLONASS COD/PHS/BIS",
            "LEAP SECONDS",
            "# OF SATELLITES",
            "PRN / # OF OBS",
            "END OF HEADER",
        ]

        head, body, i_end_head = self.get_header_body(True)
        try:
            head_sort = sorted(head, key=lambda x: header_order.index(x[60:].strip()))
            self.rinex_data = head_sort + body
        except:
            logger.warning("unable to sort header's lines, action skipped (RNXv3 only)")
        return


# *****************************************************************************
# low level functions (outside RinexFile class)
def search_idx_value(data, field):
    """
    find the index (line number) of a researched field in the RINEX data
    return None if nothing has beeen found
    """
    idx = -1
    out_idx = None
    for e in data:
        idx += 1
        if field in e:
            out_idx = idx
            break
    return out_idx


def slice_list(seq, num):
    """make sublist of num elts of a list"""
    # http://stackoverflow.com/questions/4501636/creating-sublists
    return [seq[i : i + num] for i in range(0, len(seq), num)]


def round_time(dt=None, date_delta=timedelta(minutes=60), to="average"):
    """
    Round a datetime object to a multiple of a timedelta
    dt : datetime.datetime object, default now.
    dateDelta : timedelta object, we round to a multiple of this, default 1 minute.
    to : up / down / average
    from:  http://stackoverflow.com/questions/3463930/how-to-round-the-minute-of-a-datetime-object-python
    """
    round_to = date_delta.total_seconds()
    if dt is None:
        dt = datetime.now()
    seconds = (dt - dt.min).seconds

    if seconds % round_to == 0 and dt.microsecond == 0:
        rounding = (seconds + round_to / 2) // round_to * round_to
    else:
        if to == "up":
            # // is a floor division, not a comment on following line (like in javascript):
            rounding = (
                (seconds + dt.microsecond / 1000000 + round_to) // round_to * round_to
            )
        elif to == "down":
            rounding = seconds // round_to * round_to
        else:
            rounding = (seconds + round_to / 2) // round_to * round_to

    return dt + timedelta(0, rounding - seconds, -dt.microsecond)


def regex_pattern_rinex_filename():
    """
    return a dictionnary with the different REGEX patterns to describe a RIENX filename
    """
    pattern_dic = dict()
    # pattern_dic["shortname"] = "....[0-9]{3}(\d|\D)\.[0-9]{2}(o|d)(|\.(Z|gz))"
    pattern_dic["shortname"] = (
        r"....[0-9]{3}(\d|\D)([0-9]{2}\.|\.)[0-9]{2}(o|d)(|\.(Z|gz))"  ### add subhour starting min
    )
    pattern_dic["longname"] = (
        r".{4}[0-9]{2}.{3}_(R|S|U)_[0-9]{11}_([0-9]{2}\w)_[0-9]{2}\w_\w{2}\.\w{3}(\.gz|)"
    )
    pattern_dic["longname_gfz"] = (
        r".{4}[0-9]{2}.{3}_[0-9]{8}_.{3}_.{3}_.{2}_[0-9]{8}_[0-9]{6}_[0-9]{2}\w_[0-9]{2}\w_[A-Z]*\.\w{3}(\.gz)?"
    )

    return pattern_dic


def dates_from_rinex_filename(rnx_inp):
    """
    determine the start epoch, the end epoch and the period of a RINEX
    file based on its name only.
    The RINEX is not readed. This function is much faster but less reliable
    than the RinexFile.start_date and RinexFile.end_date attribute

    return the start epoch and end epoch as datetime
    and the period as timedelta
    """
    pattern_dic = regex_pattern_rinex_filename()

    pattern_shortname = re.compile(pattern_dic["shortname"])
    pattern_longname = re.compile(pattern_dic["longname"])
    pattern_longname_gfz = re.compile(pattern_dic["longname_gfz"])

    rinexname = os.path.basename(rnx_inp)

    def _period_to_timedelta(peri_inp):
        peri_val = int(peri_inp[0:2])
        peri_unit = str(peri_inp[2])

        if peri_unit == "M":
            unit_sec = 60
        elif peri_unit == "H":
            unit_sec = 3600
        elif peri_unit == "D":
            unit_sec = 86400
        else:
            logger.warn("odd RINEX period: %s, assume it as 01D", peri_inp)
            unit_sec = 86400

        return timedelta(seconds=peri_val * unit_sec)

    ##### LONG rinex name
    if re.search(pattern_longname, rinexname):
        date_str = rinexname.split("_")[2]
        period_str = rinexname.split("_")[3]

        yyyy = int(date_str[:4])
        doy = int(date_str[4:7])
        hh = int(date_str[7:9])
        mm = int(date_str[9:11])
        dt_srt = datetime(yyyy, 1, 1) + timedelta(
            days=doy - 1, seconds=hh * 3600 + mm * 60
        )
        period = _period_to_timedelta(period_str)
        dt_end = dt_srt + period

        return dt_srt, dt_end, period

    ##### LONG rinex name -- GFZ's GODC internal name
    elif re.search(pattern_longname_gfz, rinexname):
        date_str = rinexname.split("_")[5]
        time_str = rinexname.split("_")[6]
        period_str = rinexname.split("_")[7]

        yyyy = int(date_str[:4])
        mo = int(date_str[4:6])
        dd = int(date_str[6:8])

        hh = int(time_str[0:2])
        mm = int(time_str[2:4])
        ss = int(time_str[4:6])

        dt_srt = datetime(yyyy, mo, dd, hh, mm, ss)
        period = _period_to_timedelta(period_str)
        dt_end = dt_srt + period

        return dt_srt, dt_end, period

    ##### SHORT rinex name
    elif re.search(pattern_shortname, rinexname):
        alphabet = list(string.ascii_lowercase)

        doy = int(rinexname[4:7])
        yy = int(rinexname[9:11])

        if yy > 80:
            year = yy + 1900
        else:
            year = yy + 2000

        if rinexname[7] in alphabet:
            h = alphabet.index(rinexname[7])
            period = timedelta(seconds=3600)
        else:
            h = 0
            period = timedelta(seconds=86400)

        dt_srt = datetime(year, 1, 1) + timedelta(days=doy - 1, seconds=h * 3600)
        dt_end = dt_srt + period
        return dt_srt, dt_end, period

    else:
        logger.error("%s has not a RINEX name well formatted", rinexname)
        return None, None, None


def file_period_from_timedelta(start_date, end_date):
    """
    return the RINEX file period (01H, 01D, 15M...) based on a
    start and end date

    Parameters
    ----------

    Returns
    -------
    file_period : str
        file period (01H, 01D, 15M...)
    session : bool
        True if the timedelta refers to a session (<01D)
        False otherwise (01D).

    """
    rndtup = lambda x, t: round_time(x, timedelta(minutes=t), "up")
    rndtdown = lambda x, t: round_time(x, timedelta(minutes=t), "down")
    rndtaver = lambda x, t: round_time(x, timedelta(minutes=t), "average")
    # rounded at the hour
    # maximum and average delta between start and end date
    delta_max = rndtup(end_date, 60) - rndtdown(start_date, 60)
    delta_ave = rndtaver(end_date, 60) - rndtaver(start_date, 60)

    hours_ave = int(delta_ave.total_seconds() / 3600)
    delta_sec = (end_date - start_date).total_seconds()

    # first, the special case : N *full* hours
    if delta_max <= timedelta(seconds=86400 - 3600) and hours_ave > 0:  ## = 23h max
        # delta_ave is a more precise delta than delta_max (average)
        file_period = str(hours_ave).zfill(2) + "H"
        session = True
    # more regular cases : 01H, 01D, nnM, or Unknown
    elif delta_max <= timedelta(seconds=3600):
        # Here we consider sub hourly cases
        session = True
        file_period = None
        for m in [5, 10, 15, 20, 30]:
            if (m * 60 - 1) <= delta_sec <= (m * 60 + 1):
                file_period = str(m).zfill(2) + "M"
        if not file_period:
            # NB: this test is useless, it is treated by the previous test
            file_period = "01H"
    elif hours_ave == 0 and delta_max > timedelta(seconds=3600):  # Note 2
        hours_max = int(delta_max.total_seconds() / 3600)
        file_period = str(hours_max).zfill(2) + "H"
        session = True
    elif (
        timedelta(seconds=3600) < delta_max <= timedelta(seconds=86400 + 3600)
    ):  # Note1
        file_period = "01D"
        session = False
    else:
        file_period = "00U"
        session = False
    # Note1: a tolerance of +/- 1 hours is given because old ashtech RINEXs
    #        includes the epoch of the next hour/day
    #        and then the present delta_max value reach 25
    #        it justifies also the necessity of the delta_ave variable

    # Note 2: very rare (but possible) case --' :
    #         very short file, riding in between* two hours ("a cheval sur 2 heures")
    #         met e.g. for "2024-08-13 11:54:00" > "2024-08-13 12:04:00"
    #         then the delta_max is 2H and the delta_ave is 0
    #         and we must introduce hours_max rather than hours_ave

    return file_period, session



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 26 14:45:44 2024

@author: psakic

logger for RinexMod
"""

import logging
import colorlog

# *****************************************************************************
# logger definition


def logger_define(level_prompt, logfile=None, level_logfile=None):
    """
    This function manage logging levels. It has two outputs, one to the prompt,
    the other to a logfile defined by 'logfile'.
    """

    logger_out = colorlog.getLogger('rinexmod')
    logger_out.propagate = True
    logger_out.setLevel(level_prompt)

    # This handler is for prompt (console)
    prompthandler = colorlog.StreamHandler()
    promptformatter = colorlog.ColoredFormatter(
        "%(asctime)s.%(msecs)03d|%(log_color)s%(levelname).1s%(reset)s|%(log_color)s%(funcName)-15s%(reset)s|%(message)s",
        datefmt="%y%m%dT%H:%M:%S",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
    )
    prompthandler.setFormatter(promptformatter)
    prompthandler.setLevel(level_prompt)
    if not len(logger_out.handlers):
        logger_out.addHandler(prompthandler)

    # This handler will write to a log file
    if logfile:
        if not level_logfile:
            level_logfile = level_prompt
        filehandler = logging.FileHandler(logfile, mode="a", encoding="utf-8")
        fileformatter = logging.Formatter(
            "%(asctime)s.%(msecs)03d|%(levelname).1s|%(funcName)-15s|%(message)s",
            datefmt="%y%m%dT%H:%M:%S",
        )
        filehandler.setFormatter(fileformatter)
        filehandler.setLevel(level_logfile)
        logger_out.addHandler(filehandler)

    return logger_out


logger = logger_define("INFO")


def logger_tester():
    logger.debug("debug message")
    logger.info("info message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
API functions of RinexMod

Created on Wed Mar  8 12:14:54 2023

@author: psakic
"""

import argparse
import multiprocessing as mp
import os
import re
import subprocess
from datetime import datetime

import hatanaka
import pandas as pd

import rinexmod as rimo
import rinexmod.gamit_meta as rimo_gmm
import rinexmod.logger as rimo_log
import rinexmod.metadata as rimo_mda
import rinexmod.rinexfile as rimo_rnx

logger = rimo_log.logger_define("INFO")


# *****************************************************************************
# define Python user-defined exceptions
class RinexModError(Exception):
    pass


class RinexModInputArgsError(RinexModError):
    pass


class RinexFileError(RinexModError):
    pass


class MetaDataError(RinexModError):
    pass


class ReturnListError(RinexModError):
    pass


# *****************************************************************************
# misc functions


def listfiles(directory, extension, recursive=True):
    # returns list of paths
    liste = []
    extension = extension.lower()
    for dirpath, dirnames, files in os.walk(directory):
        for name in files:
            if name.lower().endswith(extension):
                file = os.path.join(dirpath, name)
                liste.append(file)
        if not recursive:
            break
    return list(sorted(liste))


# get Git hash (to get a version number-equivalent of the RinexMod used)
def get_git_hash():
    """
    Gives the Git hash to have a tracking of the used version

    Returns
    -------
    7 characters Git hash
    """
    script_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    cmd = ["git", "--git-dir", script_path + "/.git", "rev-parse", "--short", "HEAD"]
    try:
        githash = (
            subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
            .decode("ascii")
            .strip()[:7]
        )
    except Exception:
        # logger.warn("unable to get the git commit version")
        githash = "xxxxxxx"

    ####NB: 2msec to run this fuction
    return githash


def make_site_id9(site_id_inp):
    """
    Converts a site ID to a 9-character format.

    This function takes a site ID and converts it to a 9-character format.
    If the input site ID is already 9 characters long, it returns the uppercase version of the input.
    If the input site ID is 4 characters long, it appends '00XXX' to the uppercase version of the input.
    Otherwise, it takes the first 4 characters of the input, converts them to uppercase, and appends '00XXX'.

    Parameters
    ----------
    site_id_inp : str
        The input site ID to be converted.

    Returns
    -------
    str
        The site ID in 9-character format.
    """
    if len(site_id_inp) == 9:
        return site_id_inp.upper()
    elif len(site_id_inp) == 4:
        return site_id_inp.upper() + "00XXX"
    else:
        return site_id_inp[:4].upper() + "00XXX"


# *****************************************************************************
# Metadata import


def metadata_input_manage(sitelog_inp, force=False):
    """
    Manage the multiple types possible for metadata inputs
    Return a list of MetaData to be handeled by find_mda4site

    Possible inputs are:
     * list of string (sitelog file paths),
     * single string (single sitelog file path)
     * single string (directory containing the sitelogs)
     * list of MetaData objects
     * single MetaData object

    Parameters
    ----------
    sitelog_inp : list of str, str, list of MetaData. single MetaData
        Various Input sitelogs.
    force : bool
        Force a single sitelog file path.

    Returns
    -------
        list of MetaData objects.
        (can be a singleton)

    """
    # single MetaData object
    if isinstance(sitelog_inp, rimo_mda.MetaData):
        return [sitelog_inp]
    # list of MetaData objects
    elif isinstance(sitelog_inp, list) and isinstance(
        sitelog_inp[0], rimo_mda.MetaData
    ):
        return sitelog_inp
    # single string or list of string
    elif isinstance(sitelog_inp, str) or isinstance(sitelog_inp, list):
        return sitlgs2mda_objs(
            sitelog_inp, force=force, return_list_even_if_single_input=True
        )
    else:
        errmsg = "Wrong Input,"
        errmsg = errmsg + "must be a list of string (path), a single string (path),"
        errmsg = errmsg + "a MetaData object, or a list of MetaData objects."
        errmsg = errmsg + f"Input given: {sitelog_inp}/{type(sitelog_inp)}"
        logger.error(errmsg)
        raise RinexModInputArgsError(errmsg)


def gamit2mda_objs(
    station_info_inp,
    lfile_inp=None,
    force_fake_coords=False,
    ninecharfile_inp=None,
    rev=False,
):
    """
    Read a GAMIT files and convert their content to MetaData objects

    Parameters
    ----------
    station_info_inp : str or pd.DataFrame
        Path of a GAMIT station.info file to obtain
        GNSS site metadata information.
    lfile_inp : str or pd.DataFrame
        Path of a GAMIT apriori apr/L-File to obtain
        GNSS site position and DOMES information.
    force_fake_coords : bool, optional
        hen using GAMIT station.info metadata without apriori coordinates in
        the L-File, gives fake coordinates at (0°,0°) to the site.
        The default is False.
    ninecharfile_inp : str, optional
        Path of a file that contains 9-char. site names.
        The default is None.
    rev : bool, optional
        reverse order of inputs for station.info filled from the
        newest to the oldest change
        (Automatic sort of the station.info file as DataFrame
         is then disabled)
        The default is False.

    Returns
    -------
    mdaobjs_lis : list
        list of MetaData objects.

    """
    if isinstance(station_info_inp, pd.DataFrame):
        df_stinfo_raw = station_info_inp
        stinfo_name = "station.info"
    else:
        df_stinfo_raw = rimo_gmm.read_gamit_station_info(
            station_info_inp, sort=False
        )  # sort = not rev if rev, no sort
        stinfo_name = os.path.basename(station_info_inp)

    if not lfile_inp:
        df_apr = pd.DataFrame(columns=["site"])
        logger.warning(
            "No L-File provided, fake coordinates will be used! "
            "(force_fake_coords forced to True)"
        )
        force_fake_coords = True
    elif isinstance(lfile_inp, pd.DataFrame):
        df_apr = lfile_inp
    else:
        df_apr = rimo_gmm.read_gamit_apr_lfile(lfile_inp)

    if not ninecharfile_inp is None:
        nine_char_dict = rimo.rinexmod_api.read_ninecharfile(ninecharfile_inp)
    else:
        nine_char_dict = dict()

    sites_isin = df_stinfo_raw["site"].isin(df_apr["site"])
    ### for the stats only
    sites_uniq = pd.Series(df_stinfo_raw["site"].unique())
    sites_isin_uniq = sites_uniq.isin(df_apr["site"].unique())
    n_sites_notin = len(sites_uniq) - sum(sites_isin_uniq)

    if n_sites_notin > 0 and not force_fake_coords:
        logger.warning(
            "%i/%i sites in %s are not in apr/lfile. They are skipped (you can force fake coords with -fc)",
            n_sites_notin,
            len(sites_uniq),
            stinfo_name,
        )
        df_stinfo = df_stinfo_raw[sites_isin]
    elif n_sites_notin > 0 and force_fake_coords:
        logger.warning(
            "%i/%i sites in %s are not in apr/lfile. Fake coords at (0°,0°) used",
            n_sites_notin,
            len(sites_uniq),
            stinfo_name,
        )
        df_stinfo = df_stinfo_raw
    else:  #### no missing coords, n_sites_notin == 0
        df_stinfo = df_stinfo_raw

    df_stinfo_grp = df_stinfo.groupby("site")

    mdaobjs_lis = []

    logger.info("%i sites will be extracted from %s", len(df_stinfo_grp), stinfo_name)

    for site, site_info in df_stinfo_grp:
        logger.debug("extract %s from %s", site, stinfo_name)

        #### search in ninecharfile IMPROVE ME !!!
        if site in nine_char_dict.keys():
            site_use = nine_char_dict[site]
            logger.debug("4 > 9 char. conversion: %s > %s", site, site_use)
        else:
            site_use = site

        mdaobj = rimo_mda.MetaData(sitelogfile=None)
        mdaobj.set_from_gamit(
            site_use,
            df_stinfo,
            df_apr,
            force_fake_coords=force_fake_coords,
            station_info_name=stinfo_name,
        )
        if rev:
            mdaobj.instrus.reverse()
            logger.info("Reversing order of instrumental changes")
        mdaobjs_lis.append(mdaobj)

    logger.info("%i sites have been extracted from %s", len(mdaobjs_lis), stinfo_name)

    return mdaobjs_lis


def sitlgs2mda_objs(
    sitelog_filepath, force=False, return_list_even_if_single_input=True
):
    """
    Read a set of sitelog files and convert them to MetaData objects

    Parameters
    ----------
    sitelog_filepath : str or list of str
        path of a single sitelog file or a set of sitelogs (stored in a list).
    force : bool, optional
        if it is not possible to read a sitelog (bad format), we skip it.
        The default is False.
    return_list_even_if_single_input : bool, optional
        if a single sitelog is given, return the corresponding MetaData object
        into a list (singleton). The default is True.

    Raises
    ------
    MetaDataError
    RinexModInputArgsError

    Returns
    -------
    mdaobjs_lis : list
        list of MetaData objects.

    """

    #### differenciating cases
    # Case of a list of sitelogs
    if type(sitelog_filepath) is list:
        slgs_all = sitelog_filepath
        sitelog_filepath = "input sitelog list"
    # Case of one single sitelog
    elif os.path.isfile(sitelog_filepath):
        slgs_all = [sitelog_filepath]
    # Case of a folder
    elif os.path.isdir(sitelog_filepath):
        sitelog_extension = ".log"
        slgs_all = listfiles(sitelog_filepath, sitelog_extension)

        sitelog_pattern = re.compile(r"\w{4,9}_\d{8}.log")
        slgs_all = [f for f in slgs_all if sitelog_pattern.match(os.path.basename(f))]
    ### case of no file nor folder
    else:
        errmsg = "unable to handle file/directory. Does it exists?:" + sitelog_filepath
        logger.error(errmsg)
        raise RinexModInputArgsError(errmsg)

    #### Read the sitelogs
    logger.info("**** %i sitelogs detected (in %s)", len(slgs_all), sitelog_filepath)
    for sl in slgs_all:
        logger.debug(sl)
    # Get last version of sitelogs if multiple available
    # (1st gross search based on date in filename)
    slgs_latest = _slg_find_latest_name(slgs_all)
    # load the sitelogs as metadata objects
    mdaobjs_lis, slgs_bad_lis = load_sitelogs(slgs_latest, force)
    # Get last version of sitelogs if multiple available
    # (2nd fine search based on date in "Date Prepared" field)
    mdaobjs_lis = _mda_find_latest_prep(mdaobjs_lis)

    logger.info("**** %i most recent sitelogs selected", len(mdaobjs_lis))

    for sl in mdaobjs_lis:
        logger.debug(sl.path)

    if len(slgs_bad_lis) > 0:
        logger.warning("**** %i badly-parsed & ignored sitelogs", len(slgs_bad_lis))

    if len(mdaobjs_lis) <= 1 and not return_list_even_if_single_input:
        mdaobjs_lis = mdaobjs_lis[0]

    return mdaobjs_lis


def rinexs2mda_objs(rinex_paths, ninecharfile_inp=None):
    """
    Read a set of RINEX files and convert them to MetaData objects

    Parameters
    ----------
    rinex_paths : list of str
        path of a single rinex file or a set of rinex files (stored in a list).

    ninecharfile_inp : str, optional
        Path of a file that contains 9-char. site names.
        The default is None.

    Returns
    -------
    mdaobjs_lis : list
        list of MetaData objects.

    """

    if not isinstance(rinex_paths, list):
        rinex_paths = [rinex_paths]

    if not ninecharfile_inp is None:
        nine_char_dict = read_ninecharfile(ninecharfile_inp)
    else:
        nine_char_dict = dict()

    mda_stk = []
    for rnx in sorted(rinex_paths):
        mda = rimo_mda.MetaData()
        mda.set_from_rinex(rnx)

        #### search in ninecharfile IMPROVE ME !!!
        if mda.site_id9[-3:] == "XXX" and mda.site_id4 in nine_char_dict.keys():
            site_id4_orig = mda.site_id4
            mda.site_id = nine_char_dict[mda.site_id4]
            logger.info(f"4 > 9 char. conversion: {site_id4_orig:} > {mda.site_id9:}")

        mda_stk.append(mda)

    mdaobjs_lis, mdaobj_dic = group_mda(mda_stk)

    return mdaobjs_lis


def group_mda(mdaobj_list_inp):
    """
    Group a list of MetaData objects into a dictionary and return a list and dictionary of merged objects.

    This function processes a list of MetaData objects, grouping them by their `site_id9` attribute.
    If multiple MetaData objects share the same `site_id9`, their instrumentation data is merged.

    Parameters
    ----------
    mdaobj_list_inp : list
        A list of MetaData objects to be grouped.

    Returns
    -------
    mdaobj_lis : list
      A list of grouped MetaData objects.
    mdaobj_dic : dict
        A dictionary where keys are `site_id9` and values are the grouped MetaData objects.
    """
    mdaobj_dic = (
        dict()
    )  # Initialize an empty dictionary to store grouped MetaData objects.
    for mda in mdaobj_list_inp:
        # Check if the current MetaData object's site_id9 is already in the dictionary.
        if not mda.site_id9 in mdaobj_dic.keys():
            # If not, add it to the dictionary.
            mdaobj_dic[mda.site_id9] = mda
        else:
            # If it exists, merge the instrumentation data from the current object.
            for inst in mda.instrus:
                mdaobj_dic[mda.site_id9].add_instru(
                    inst["receiver"],
                    inst["antenna"],
                    inst["dates"][0],
                    inst["dates"][1],
                )

    # Convert the dictionary values to a list of grouped MetaData objects.
    mdaobj_lis = [v for k, v in mdaobj_dic.items()]

    return mdaobj_lis, mdaobj_dic


def load_sitelogs(sitelogs_inp, force=False):
    """
    Process a list of sitelogs and return a list of MetaData objects and a list of bad sitelogs.

    Parameters
    ----------
    sitelogs_inp : list
        List of paths to the latest sitelog files.
    force : bool, optional
        If True, force processing even if sitelog is not parsable. Default is False.

    Returns
    -------
    mdaobjs_lis : Iterable
        List of MetaData objects.
    bad_sitelogs_lis : Iterable
        List of sitelogs that could not be parsed.
    """

    def _load_slg(sitelog_filepath, force_load):
        # Creating MetaData object
        try:
            mdaobj_load = rimo_mda.MetaData(sitelog_filepath)
        except Exception as e:
            # If sitelog is not parsable
            errmsg = f"The sitelog is not parsable: {os.path.basename(sitelog_filepath)} ({str(e)})"
            logger.error(errmsg)
            if not force_load:
                raise MetaDataError(errmsg)
            else:
                mdaobj_load = None

        return mdaobj_load

    mdaobjs_lis = []
    bad_sitelogs_lis = []
    for sta_sitelog in sitelogs_inp:
        mdaobj = _load_slg(sta_sitelog, force)
        # Appending to list
        if mdaobj:
            mdaobjs_lis.append(mdaobj)
        else:
            bad_sitelogs_lis.append(sta_sitelog)

    return mdaobjs_lis, bad_sitelogs_lis


def _slg_find_latest_name(all_sitelogs_filepaths):
    """
    Find the latest version of a sitelog within a list of sitelogs,
    based on date in its filename
    (mainly for time consumption reduction)

    see also _mda_find_latest_prep (more reliable but slower)
    """
    # We list the available sites to group sitelogs
    bnm = os.path.basename
    sl_bnm = [bnm(sl) for sl in all_sitelogs_filepaths]
    sl_sta = [sl[0:4] for sl in sl_bnm]

    # set the output latest sitelog list
    latest_sitelogs_filepaths = []

    for sta in sl_sta:
        # Grouping by site
        sta_sitelogs = [
            slp for (slp, sln) in zip(all_sitelogs_filepaths, sl_bnm) if sln[0:4] == sta
        ]
        # Getting dates from basename
        # sitelogs_dates0 = [os.path.splitext(bnm(sl))[0][-8:] for sl in sta_sitelogs]
        # Getting dates from basename and parsing 'em
        date_from_fn = lambda x: os.path.splitext(bnm(x))[0][-8:]
        sitelogs_dates = []
        for sl in sta_sitelogs:
            try:
                d = datetime.strptime(date_from_fn(sl), "%Y%m%d")
                sitelogs_dates.append(d)
            except ValueError as e:
                errmsg = f"bad date {date_from_fn(sl):} in sitelog's filename: {sl:}"
                logger.error(errmsg)
                raise e
        # We get the max date and put it back to string format.
        maxdate = max(sitelogs_dates).strftime("%Y%m%d")
        # We filter the list with the max date string, and get a one entry list, then transform it to string
        sta_sitelog = [sl for sl in sta_sitelogs if maxdate in date_from_fn(sl)][0]

        latest_sitelogs_filepaths.append(sta_sitelog)

    return latest_sitelogs_filepaths


def _mda_find_latest_prep(mdaobjs_inp):
    """
    Find the latest version of a MetaData object within a list of MetaData objects,
    based on preparation date

    more reliable than _slg_find_latest_name but requires that the metadata objects have been loaded
    """
    # We list the available sites to group sitelogs
    mdaobjs_lis = mdaobjs_inp
    # mdaobjs_lis = sorted(mdaobjs_lis, key=lambda x: x.misc_meta["date prepared"], reverse=True)

    # set the output latest sitelog list
    mdaobjs_latest = []

    sites_all = list(sorted(list(set([md.site_id4 for md in mdaobjs_lis]))))

    for site in sites_all:
        # Grouping by site
        mdaobjs_site = [m for m in mdaobjs_lis if m.site_id4 == site]
        # Getting dates from basename and parsing 'em
        mdaobjs_site_dates = [m.misc_meta["date prepared"] for m in mdaobjs_site]
        # We get the max date and put it back to string format.
        maxdate = max(mdaobjs_site_dates)
        # We filter the list with the max date string, and get a one entry list, then transform it to string
        mdaobj_latest = [
            md for md in mdaobjs_site if md.misc_meta["date prepared"] == maxdate
        ][0]

        mdaobjs_latest.append(mdaobj_latest)

    return mdaobjs_latest


def find_mda4site(rnxobj_or_site4char, mdaobjs_lis, force):
    """
    Finding the right MetaData object

    If is list, can not use force.
    If no metadata found, do not process.
    """
    if type(rnxobj_or_site4char) is str:
        rnx_4char = rnxobj_or_site4char[:4]
        ermg_fna = rnxobj_or_site4char
    else:
        rnx_4char = rnxobj_or_site4char.get_site(True, True)
        ermg_fna = rnxobj_or_site4char.filename

    logger.debug("Searching corresponding metadata for site: " + rnx_4char)

    if len(mdaobjs_lis) == 0:
        logger.warning("The metadata list provided is empty!")

    mdaobj = None
    if rnx_4char not in [mda.site_id4 for mda in mdaobjs_lis]:
        if len(mdaobjs_lis) == 1:
            if not force:
                ermg_lbl = "RINEX name's site does not correspond to provided metadata, use -f option to force"
                errmsg = f"{ermg_lbl:110s} - {ermg_fna:}"
                logger.error(errmsg)
                raise RinexModInputArgsError(errmsg)
            else:
                ermg_lbl = "RINEX name's site does not correspond to provided metadata, forced processing anyway"
                logger.warning(f"{ermg_lbl:110s} - {ermg_fna:}")
        else:
            ermg_lbl = "No metadata found for this RINEX"
            errmsg = f"{ermg_lbl:110s} - {ermg_fna:}"
            logger.error(errmsg)
            raise RinexModInputArgsError(errmsg)
    else:
        mdaobjs_site = [md for md in mdaobjs_lis if md.site_id4 == rnx_4char]
        if len(mdaobjs_site) == 1:
            mdaobj = mdaobjs_site[0]
        else:
            # the assumption that latest sitelog has been found in sitlgs2mda_objs is wrong!
            # a second search with the preparation date is performed here
            mdaobj = _mda_find_latest_prep(mdaobjs_site)[0]

    return mdaobj


def apply_mda2rnxobj(rnxobj, mdaobj, ignore=False, keep_rnx_rec=False):
    """
    apply a MetaData object on a RinexFile object
    to modify this RinexFile with the rights metadata
    """

    ### do this check with 9 chars at one point
    rnx_4char = rnxobj.get_site(True, True)
    # Site name from the sitelog
    mda_4char = mdaobj.site_id4  # misc_meta["ID"].lower()[:4]

    if rnx_4char != mda_4char:
        logger.warning(
            "RINEX and metadata 4 char. codes do not correspond, "
            "but I assume you know what you are doing (%s,%s)",
            rnx_4char,
            mda_4char,
        )

    # Get rinex header values from sitelog infos and start and end time of the file
    # ignore option is to ignore firmware changes between instrumentation periods.
    mda_vars, ignored = mdaobj.find_instru4rnx(
        rnxobj.start_date, rnxobj.end_date, ignore
    )

    if not mda_vars:
        errmsg = f"No instrumentation corresponding to the RINEX epoch - {rnxobj.filename}"
        logger.error(errmsg)
        raise MetaDataError(errmsg)

    if ignored:
        logger.warning(
            "Instrumentation comes from merged metadata periods with different firmwares, processing anyway - %s",
            rnxobj.filename,
        )

    (
        fourchar_id,
        domes_id,
        sat_sys_long_fmt,
        agencies,
        rec,
        antenna,
        ant_pos,
        ant_delta,
    ) = mda_vars

    ## Apply the modifications to the RinexFile object
    rnxobj.mod_marker(fourchar_id, domes_id)
    rnxobj.mod_receiver(keep_rnx_rec=keep_rnx_rec, **rec)
    rnxobj.mod_interval(rnxobj.sample_rate_numeric)
    rnxobj.mod_antenna(**antenna)
    rnxobj.mod_antenna_pos(**ant_pos)
    rnxobj.mod_antenna_delta(**ant_delta)
    rnxobj.mod_agencies(**agencies)
    if not keep_rnx_rec:
        ### if keep_rnx_rec is active, we keep the sat system in the header
        rnxobj.mod_sat_system(sat_sys_long_fmt)

    return rnxobj


# *****************************************************************************
# modification keyword dictionary functions


def _modif_kw_check(modif_kw):
    """
    Check if acceptable modification keywords have been provided

    Raise a RinexModInputArgsError Exception if not
    """

    acceptable_keywords = [
        "station",
        "marker_name",
        "marker_number",
        "receiver_serial",
        "receiver_type",
        "receiver_fw",
        "antenna_serial",
        "antenna_type",
        "antenna_X_pos",
        "antenna_Y_pos",
        "antenna_Z_pos",
        "antenna_H_delta",
        "antenna_E_delta",
        "antenna_N_delta",
        "operator",
        "agency",
        "sat_system",
        "observables",
        "interval",
        "filename_data_freq",
        "filename_file_period",
        "filename_data_source",
        "comment(_[0-9]+)?",
    ]
    ### comment is a regex, bc several comments are possible
    # suffix _N is added by ParseKwargs
    # but comment without suffix must remain possible (for API mode)

    for kw in modif_kw:
        if not any([re.match(akw, kw) for akw in acceptable_keywords]):
            errmsg = f"'{kw}' is not an acceptable keyword for header modification."
            logger.error(errmsg)
            return RinexModInputArgsError(errmsg)

    return None


def apply_modifkw2rnxobj(rnxobj, modif_kw):
    """
    apply a modification keywords on a RinexFile object
    to modify this RinexFile with the rights metadata
    """

    def __keys_in_modif_kw(keys_in):
        return all([e in modif_kw.keys() for e in keys_in])

    rnxobj.mod_marker(modif_kw.get("marker_name"), modif_kw.get("marker_number"))

    # legacy keyword, 'marker_name' should be used instead
    rnxobj.mod_marker(modif_kw.get("station"))

    rnxobj.mod_receiver(
        modif_kw.get("receiver_serial"),
        modif_kw.get("receiver_type"),
        modif_kw.get("receiver_fw"),
    )

    rnxobj.mod_antenna(modif_kw.get("antenna_serial"), modif_kw.get("antenna_type"))

    rnxobj.mod_antenna_pos(
        modif_kw.get("antenna_X_pos"),
        modif_kw.get("antenna_Y_pos"),
        modif_kw.get("antenna_Z_pos"),
    )

    rnxobj.mod_antenna_delta(
        modif_kw.get("antenna_H_delta"),
        modif_kw.get("antenna_E_delta"),
        modif_kw.get("antenna_N_delta"),
    )

    rnxobj.mod_agencies(modif_kw.get("operator"), modif_kw.get("agency"))

    rnxobj.mod_sat_system(modif_kw.get("sat_system"))
    # legacy keyword, 'sat_system' should be used instead
    rnxobj.mod_sat_system(modif_kw.get("observables"))

    rnxobj.mod_interval(modif_kw.get("interval"))

    # for the filename
    rnxobj.mod_filename_file_period(modif_kw.get("filename_file_period"))
    rnxobj.mod_filename_data_freq(modif_kw.get("filename_data_freq"))
    rnxobj.mod_filename_data_source(modif_kw.get("filename_data_source"))

    # comment
    # special case: several keys comment_1, comment_2, comment_N are possible
    # number are added automatically by ParseKwargs
    comment_keys = [k for k in modif_kw.keys() if "comment" in k]
    for ck in comment_keys:
        rnxobj.add_comment(modif_kw.get(ck).split("_")[0])

    return rnxobj


# *****************************************************************************
# dictionnary as output for gnss_delivery workflow


def _return_lists_maker(rnxobj_or_dict, return_lists=dict()):
    """
    Construct the so called ``return_lists`` (which are actually dictionnaries)

    return_lists have the structure:

    ```
    return_lists[major_rinex_version][sample_rate_string][file_period]
    ```

    the input ``return_lists`` is populated with the content of the input
    ``rnxobj_or_dict``

    Parameters
    ----------
    rnxobj_or_dict : RinexFile object or dict

        The input new information which will populate the return_lists

        if a dict is provided, we assume it is a singleton with a standard
        return_lists structure to be merged in a bigger return_lists

    return_lists : dict, optional
        A potential pre-exisiting return_lists.
        Per default it is a brand-new return_lists for scratch.
        The default is dict().

    Returns
    -------
    return_lists : dict
        The input return_lists populated with the input rnxobj_or_dict.

    Note
    ----
    Specific usage for the IPGP's gnss_delivery workflow

    """

    if isinstance(rnxobj_or_dict, rimo_rnx.RinexFile):
        rnxobj = rnxobj_or_dict
        major_rinex_version = rnxobj.version[0]
        sample_rate_string = rnxobj.sample_rate_string
        file_period = rnxobj.file_period
        path_output = rnxobj.path_output
    elif isinstance(rnxobj_or_dict, dict):
        rtrnlst = rnxobj_or_dict
        major_rinex_version = list(rtrnlst.keys())[0]
        sample_rate_string = list(rtrnlst[major_rinex_version].keys())[0]
        file_period = list(rtrnlst[major_rinex_version][sample_rate_string].keys())[0]
        path_output = rtrnlst[major_rinex_version][sample_rate_string][file_period][0]
    else:
        errmsg = f"Wrong Input, must be RinexFile object or dict. Input given: {rnxobj_or_dict}/{type(rnxobj_or_dict)}"
        logger.error(errmsg)
        raise ReturnListError(errmsg)

    # Dict ordered as : RINEX_VERSION, SAMPLE_RATE, FILE_PERIOD
    if major_rinex_version not in return_lists:
        return_lists[major_rinex_version] = {}
    if sample_rate_string not in return_lists[major_rinex_version]:
        return_lists[major_rinex_version][sample_rate_string] = {}
    if file_period not in return_lists[major_rinex_version][sample_rate_string]:
        return_lists[major_rinex_version][sample_rate_string][file_period] = []

    return_lists[major_rinex_version][sample_rate_string][file_period].append(
        path_output
    )

    return return_lists


def _return_lists_write(return_lists, logfolder, now_dt=None):
    # Writing an output file for each RINEX_VERSION, SAMPLE_RATE, FILE_PERIOD lists
    if not now_dt:
        now_dt = datetime.now()

    this_outputfile = ""

    for rinex_version in return_lists:
        for sample_rate in return_lists[rinex_version]:
            for file_period in return_lists[rinex_version][sample_rate]:
                this_outputfile = "_".join(
                    [
                        "RINEX" + rinex_version,
                        sample_rate,
                        file_period,
                        datetime.strftime(now_dt, "%Y%m%d%H%M"),
                        "delivery.lst",
                    ]
                )
                this_outputfile = os.path.join(logfolder, this_outputfile)

                # Writting output to temporary file and copying it them to target files
                with open(this_outputfile, "w") as f:
                    f.writelines(
                        "{}\n".format(line)
                        for line in return_lists[rinex_version][sample_rate][
                            file_period
                        ]
                    )
                    logger.debug("Output rinex list written to " + this_outputfile)

    return this_outputfile


def read_ninecharfile(ninecharfile_inp):
    nine_char_dict = dict()

    if isinstance(ninecharfile_inp, str):
        with open(ninecharfile_inp, "r") as F:
            nine_char_list = F.readlines()
    elif isinstance(ninecharfile_inp, list):
        nine_char_list = ninecharfile_inp
    else:
        nine_char_list = list(ninecharfile_inp)

    for site_key in nine_char_list:
        nine_char_dict[site_key[:4].lower()] = site_key.strip()

    return nine_char_dict


# *****************************************************************************
# Main function


def rinexmod(
    rinexfile,
    outputfolder,
    sitelog=None,
    modif_kw=dict(),
    marker="",
    country="",
    longname=False,
    shortname=False,
    force_rnx_load=False,
    force_sitelog=False,
    ignore=False,
    ninecharfile=None,
    no_hatanaka=False,
    compression="gz",
    relative="",
    verbose=True,
    full_history=False,
    filename_style="basic",
    return_lists=None,
    station_info=None,
    lfile_apriori=None,
    force_fake_coords=False,
    remove=False,
    keep_rnx_rec=False,
):
    """
    Parameters
    ----------
    rinexfile : str
        Input RINEX file to process.
    outputfolder : str
        Folder where to write the modified RINEX files.
    sitelog : str, list of str, MetaData object, list of MetaData objects, optional
        Get the RINEX header values from a sitelog.
        Possible inputs are:
         * list of string (sitelog file paths),
         * single string (single sitelog file path or directory containing the sitelogs),
         * list of MetaData object
         * single MetaData object
        The function will search for the latest and right sitelog
        corresponding to the site.
        One can force a single sitelog with force_sitelog.
        The default is None.
    modif_kw : dict, optional
        Modification keywords for RINEX's header fields and/or filename.
        Will override the information from the sitelog.
        Acceptable keywords for the header fields:
         * comment
         * marker_name
         * marker_number
         * station (legacy alias for marker_name)
         * receiver_serial
         * receiver_type
         * receiver_fw
         * antenna_serial
         * antenna_type
         * antenna_X_pos
         * antenna_Y_pos
         * antenna_Z_pos
         * antenna_H_delta
         * antenna_E_delta
         * antenna_N_delta
         * operator
         * agency
         * sat_system (M, G, R, E, C...)
         * observables (legacy alias for sat_system)
         * interval
        Acceptable keywords for the filename:
         * filename_file_period (01H, 01D...)
         * filename_data_freq (30S, 01S...)
         * filename_data_source (R, S, U)
        The default is dict().
    marker : str, optional
        A four or nine character site code that will be used to rename
        input files.
        Apply also to the header's MARKER NAME,
        but a custom modification keyword marker_name='XXXX' overrides it
        (modif_kw argument below)
        The default is ''.
    country : str, optional
        A three character string corresponding to the ISO 3166 Country code
        that will be used to rename input files.
        It overrides other country code sources (sitelog, --marker...)
        list of ISO country codes:
        https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes
        The default is ''.
    longname : bool, optional
        Force RINEX file renaming with long name convention (force gzip compression).
        Mutually exclusive with shortname.
        The default is False.
    shortname : bool, optional
        Force RINEX file renaming with short name convention.
        Mutually exclusive with longname.
        The default is False.
    force_rnx_load : bool, optional
        Force the loading of the input RINEX. Useful if its name is not standard.
        The default is False.
    force_sitelog : bool, optional
        If a single sitelog is provided, force sitelog-based header
        values when RINEX's header and sitelog site name do not correspond.
        If several sitelogs are provided, skip badly-formated sitelogs.
        The default is False.
    ignore : bool, optional
        Ignore firmware changes between instrumentation periods
        when getting header values info from sitelogs. The default is False.
    ninecharfile : str, optional
        Path of a file that contains 9-char. site names from the M3G database.
        The default is None.
    no_hatanaka : bool, optional
        Skip high-level RINEX-specific Hatanaka compression
        (performed per default).
        The default is False.
    compression : str, optional
        Set low-level RINEX file compression.
        acceptable values : 'gz' (recommended to fit IGS standards), 'Z', None.
        The default is 'gz'.
    relative : str, optional
        Reconstruct files relative subfolders.
        You have to indicate the common parent folder,
        that will be replaced with the output folder. The default is ''.
    verbose : bool, optional
        set the level of verbosity
        (False for the INFO level, True for the DEBUG level).
        The default is True.
    full_history : bool, optional
        Add the full history of the station in
        the RINEX's header as comment.
    filename_style : str, optional
        Set the RINEX filename style.
        acceptable values : 'basic' (per default), 'flex', 'exact'.
        * 'basic': a simple mode to apply a strict filename period (01H or 01D),
        being compatible with the IGS conventions.
        e.g.: `FNG000GLP_R_20242220000_01D_30S_MO.crx.gz`
        * 'flex': the filename period is tolerant and corresponds to
        the actual data content, but then can be odd (e.g. 07H, 14H...).
        The filename start time is rounded to the hour.
        e.g.: `FNG000GLP_R_20242221800_06H_30S_MO.crx.gz`
        * 'exact': the  filename start time is strictly the one of the
        first epoch in the RINEX.
        Useful for some specific cases needing splicing.
        e.g.: `FNG000GLP_R_20242221829_06H_30S_MO.crx.gz`
        The default is 'basic'.
    return_lists : dict, optional
        Specific option for file distribution through a GLASS node.
        Store the rinexmoded RINEXs in a dictionary
        to activates it, give a dict as input (an empty one - dict() works)
        The default is None.
    station_info: str, optional
        Path of a GAMIT station.info file to obtain GNSS site
        metadata information (needs also lfile_apriori option)
    lfile_apriori: str, optional
        Path of a GAMIT apriori apr/L-File to obtain GNSS site
        position and DOMES information (needs also station_info option)
    force_fake_coords: bool, optional
        When using GAMIT station.info metadata without apriori coordinates
        in the L-File, gives fake coordinates at (0°,0°) to the site
    remove: bool, optional
        Remove input RINEX file if the output RINEX is correctly written
        The default is False.
    keep_rnx_rec: bool, optional
        Keep the RINEX receiver header record in the output RINEX.
        Metadata from the external source (e.g. sitelogs) will not be modded.
        The default is False.

    Raises
    ------
    RinexModInputArgsError
        Something is wrong with the input arguments.
    RinexFileError
        Something is wrong with the input RINEX File.

    Returns
    -------
    outputfile : str
        the path of the rinexmoded RINEX

    OR

    return_lists : dict
        a dictionary of rinexmoded RINEXs for GLASS distribution.
    """

    now = datetime.now().astimezone()

    if verbose:
        logger = rimo_log.logger_define("DEBUG", logfile=None, level_logfile="DEBUG")
    else:
        logger = rimo_log.logger_define("INFO", logfile=None, level_logfile="INFO")

    logger.info("# Inp. file: %s", rinexfile)

    if relative:
        if not relative in rinexfile:
            errmsg = "{:110s} - {}".format(
                "The relative subfolder can not be reconstructed for RINEX file",
                rinexfile,
            )
            logger.error(errmsg)
            raise RinexModInputArgsError(errmsg)

        # We construct the output path with relative path between file name and parameter
        relpath = os.path.relpath(os.path.dirname(rinexfile), relative)
        myoutputfolder = os.path.join(outputfolder, relpath)
        if not os.path.isdir(myoutputfolder):
            os.makedirs(myoutputfolder)
    elif os.path.basename(outputfolder) == "INPUT_FOLDER":
        myoutputfolder = os.path.dirname(rinexfile)
    else:
        myoutputfolder = outputfolder

    if not modif_kw:
        modif_kw = dict()

    if (
        os.path.abspath(os.path.dirname(rinexfile)) == myoutputfolder
        and not outputfolder == "IDEM"
    ):
        errmsg = "{:110s} - {}".format(
            "Input and output folders are the same!", rinexfile
        )
        logger.error(errmsg)
        raise RinexFileError(errmsg)

    if outputfolder == "IDEM":
        logger.warning("The output folder is forced as the same one as the input one")

    if not os.path.exists(outputfolder):
        logger.debug("The output folder does not exists, creating it")
        os.makedirs(outputfolder)

    if longname and shortname:
        errmsg = "longname and shortname are mutually exclusive"
        logger.error(errmsg)
        raise RinexModInputArgsError(errmsg)

    ###########################################################################
    ########## Open the rinex file as an object
    rnxobj = rimo_rnx.RinexFile(rinexfile, force_rnx_load=force_rnx_load)

    if rnxobj.status:
        errmsg = "{:110s} - {}".format(rnxobj.status, rinexfile)
        logger.error(errmsg)
        raise RinexFileError(errmsg)

    logger.debug("RINEX Origin Metadata :\n" + rnxobj.get_header()[0])

    # apply tolerant / strict (per default) file period
    if filename_style == "basic":
        rnxobj.mod_file_period_basic()

    # Check that the provided marker is a 4-char site name
    if marker and (len(marker) != 4 and len(marker) != 9):
        errmsg = "The site name provided is not 4 or 9-char valid: " + marker
        logger.error(errmsg)
        raise RinexModInputArgsError(errmsg)

    # Get the 4 char > 9 char dictionnary from the input list
    nine_char_dict = dict()  # in any case, nine_char_dict is initialized
    if ninecharfile:
        if not os.path.isfile(ninecharfile):
            errmsg = "The specified 9-chars. list file does not exists: " + ninecharfile
            logger.error(errmsg)
            raise RinexModInputArgsError(errmsg)

        nine_char_dict = read_ninecharfile(ninecharfile)

    # set the marker as Rinex site, if any
    # This preliminary set_site is for th research of the right sitelog
    # a second set_site will take place a just after with more details
    if marker:
        # We store the old site name to add a comment in rinex file's header
        ## modif_marker = rnxobj.get_site(True,False) ### Useless...
        rnxobj.set_site(marker)

    ## warning if no metadata at all is not provided
    if not sitelog and not modif_kw and (not station_info or not lfile_apriori):
        logger.warning(
            "No sitelog nor keywords nor station.info+lfile provided. "
            "Per default rec.'s header will remain & no new "
            "metdata will be written!"
        )

    # initialize the metadata object list (to avoid linter warning)
    mdaobjs_lis = []
    ## sitelogs
    if sitelog:
        mdaobjs_lis = metadata_input_manage(sitelog, force=force_sitelog)
    ## GAMIT files
    # We read the GAMIT files only if no input 'sitelog' variable is given.
    # Indeed, the GAMIT files might have been read outside this function
    # (most likely actually). The already read GAMIT files are then stored
    # in the 'sitelog' variable as a list of MetaData objects

    if (station_info and not lfile_apriori) or (not station_info and lfile_apriori):
        errmsg = "station_info and lfile_apriori must be provided together"
        logger.critical(errmsg)
        raise RinexModInputArgsError(errmsg)

    ### load the metadata from sitelog or GAMIT files if any
    if (station_info and lfile_apriori) and not sitelog:
        mdaobjs_lis = gamit2mda_objs(
            station_info, lfile_apriori, force_fake_coords=force_fake_coords
        )

    ### find the right MetaData object corresponding to the RINEX
    if sitelog or (station_info and lfile_apriori):
        mdaobj = find_mda4site(rnxobj, mdaobjs_lis, force=force_sitelog)
        logger.debug("metadata used: %s", mdaobj)
    else:
        mdaobj = None

    ###########################################################################
    ########## Handle the similar options to set the site code
    ### Priority for the Country source
    # 0) if --country is given it overrides everything,
    #    but at the end of the tests
    # 1) the marker option if 9 char are given
    # 2) the nine_char_dict from the ninecharfile option
    # 3) the MetaData object (most useful actually),
    #    but we maintain a fallback mechanism here if the sitelog is wrong
    # 4) last chance: test if the country code we get from the input 9-char
    #    code is not XXX. If so, we keep it
    # Finally, set default value for the monument & country codes
    #
    rnx_4char = rnxobj.get_site(True, True)
    rnx_9char = rnxobj.get_site(False, False)

    if marker and len(marker) == 9:
        monum = marker[4:6]
        cntry = marker[6:]
    elif ninecharfile:
        if not rnx_4char in nine_char_dict:
            logger.warning(
                "Site's missing in the input 9-char. file: %s", rinexfile
            )
            monum = "00"
            cntry = "XXX"
        else:
            monum = nine_char_dict[rnx_4char].upper()[4:6]
            cntry = nine_char_dict[rnx_4char].upper()[6:]
    elif mdaobj:
        monum = "00"
        cntry = mdaobj.get_country()
    elif rnx_9char[6:] != "XXX":
        monum = rnx_9char[4:6]
        cntry = rnx_9char[6:]
    else:
        monum = "00"
        cntry = "XXX"

    if country:
        if len(country) == 3:
            cntry = country
        else:
            logger.warning(
                "Input country code is not 3 chars. RINEX will not be properly renamed: %s",
                rinexfile,
            )

    if cntry == "XXX":
        logger.warning(
            "Site's country not retrieved. RINEX will not be properly renamed: %s",
            rinexfile,
        )

    rnxobj.set_site(rnx_4char, monum, cntry)

    ###########################################################################
    ########## Remove previous comments
    if mdaobj or modif_kw or (station_info and lfile_apriori):
        rnxobj.clean_rinexmod_comments(clean_history=True)

    ###########################################################################
    ########## Apply the MetaData object on the RinexFile object
    if mdaobj:
        rnxobj = apply_mda2rnxobj(
            rnxobj, mdaobj, ignore=ignore, keep_rnx_rec=keep_rnx_rec
        )
        logger.debug("RINEX Sitelog-Modified Metadata :\n" + rnxobj.get_header()[0])
        modif_source_metadata = mdaobj.filename
    else:
        modif_source_metadata = ""

    ###########################################################################
    ########## Apply the modif_kw dictionnary on the RinexFile object
    if modif_kw:
        # Checking input keyword modification arguments
        _modif_kw_check(modif_kw)

        modif_source_kw = "keywords:" + " ".join(modif_kw.keys())
        rnxobj = apply_modifkw2rnxobj(rnxobj, modif_kw)
        logger.debug(
            "RINEX Manual Keywords-Modified Metadata:\n" + rnxobj.get_header()[0]
        )
    else:
        modif_source_kw = ""

    ###########################################################################
    ########## Apply the site as the MARKER NAME within the RINEX
    # Must be after apply_mda2rnxobj and apply_modifkw2rnxobj
    # apply only is modif_kw does not overrides it (it is the overwhelming case)
    if "marker_name" not in modif_kw.keys():
        rnxobj.mod_marker(rnxobj.get_site(False, False, True))

    ###########################################################################
    ########## Correct the first and last time obs
    rnxobj.mod_time_obs(rnxobj.start_date, rnxobj.end_date)

    ###########################################################################
    ########## Add comment in the header
    githash = get_git_hash()
    vers_num = rimo.__version__ + " " + githash[-3:]
    # rnxobj.add_comment(("RinexMod (IPGP)","METADATA UPDATE"),add_pgm_cmt=True)
    rnxobj.add_prg_run_date_comment("RinexMod " + vers_num, "METADATA UPDATE")
    rnxobj.add_comment("RinexMod / IPGP-OVS (github.com/IPGP/rinexmod)")
    rnxobj.add_comment(
        "rinexmoded on {}".format(datetime.strftime(now, "%Y-%m-%d %H:%M%z"))
    )
    if mdaobj:
        rnxobj.add_comment("rinexmoded with {}".format(modif_source_metadata))
    if modif_kw:
        rnxobj.add_comment("rinexmoded with {}".format(modif_source_kw))
    # if marker: ##### Useless...
    #    rnxobj.add_comment('filename assigned from {}'.format(modif_marker))

    ###########################################################################
    ########## Write the station history in the header
    if mdaobj and full_history:
        title = ["- - - - - - - - - - - -", "SITE FULL HISTORY"]
        rnxobj.add_comments(title + mdaobj.rinex_full_history_lines())

    ###########################################################################
    ########## Sort the header
    rnxobj.sort_header()

    ###########################################################################
    ########## we regenerate the filenames
    if shortname or (rnxobj.name_conv == "SHORT" and not longname):
        apply_longname = False
    elif longname or (rnxobj.name_conv == "LONG" and not shortname):
        apply_longname = True
    else:
        apply_longname = True

    if apply_longname:
        rnxobj.get_longname(
            inplace_set=True,
            compression="",
            filename_style=filename_style,
            data_source=rnxobj.data_source,
        )
    else:
        rnxobj.get_shortname(
            inplace_set=True, compression="", filename_style=filename_style
        )

    # NB: here the compression type must be forced to ''
    #     it will be added in the next step in write_to_path

    ###########################################################################
    ########## We convert the file back to Hatanaka Compressed Rinex

    # NB: this test is complcated, ambiguous and not very useful => disabled 2025-01-05

    # if apply_longname and not compression:
    #     # If not specified, we set compression to gz when file changed to longname
    #     output_compression = "gz"
    # elif not compression:
    #     output_compression = rnxobj.compression
    # else:
    #     output_compression = compression

    output_compression = compression

    ###########################################################################
    ########## Writing output file
    try:
        outputfile = rnxobj.write_to_path(
            myoutputfolder, compression=output_compression, no_hatanaka=no_hatanaka
        )
        logger.info("# Out. file: " + outputfile)
    except hatanaka.hatanaka.HatanakaException as e:
        errmsg ="{:110s} - {}".format("File could not be written - hatanaka exception", rinexfile)
        logger.error(errmsg)
        outputfile = None
        raise e

    if remove and os.path.isfile(outputfile):
        logger.info("Input file removed: %s", rinexfile)
        os.remove(rinexfile)

    ###########################################################################
    # Construct return dict by adding key if doesn't exists
    # and appending file to corresponding list
    if type(return_lists) is dict:
        return_lists = _return_lists_maker(rnxobj, return_lists)
        final_return = return_lists
    else:
        ###########################################################################
        # if no return dict given, return simply the path of the outputfile
        final_return = outputfile

    return final_return


# *****************************************************************************
# Upper level rinexmod for a Console run


def rinexmod_cli(
    rinexinput,
    outputfolder,
    sitelog=None,
    modif_kw=dict(),
    marker="",
    country="",
    longname=False,
    shortname=False,
    force_sitelog=False,
    force_rnx_load=False,
    ignore=False,
    ninecharfile=None,
    no_hatanaka=False,
    compression="gz",
    relative="",
    verbose=True,
    alone=False,
    output_logs=None,
    write=False,
    sort=False,
    full_history=False,
    filename_style="basic",
    multi_process=1,
    debug=False,
    station_info=None,
    lfile_apriori=None,
    force_fake_coords=False,
    remove=False,
    keep_rnx_rec=False,
):
    """
    Main function for reading a Rinex list file. It processes the list, and apply
    file name modification, command line based header modification, or sitelog-based
    header modification.

    Optimized for a CLI (in a Terminal usage) but can be used also in a
    stand-alone API mode.

    For a detailed description, check the help of the lower level
    `rinexmod` function or the help of the frontend CLI function in a Terminal
    """

    # If no longname, modif_kw and sitelog, return
    if (
        not sitelog
        and not modif_kw
        and not marker
        and not longname
        and not shortname
        and not station_info
        and not lfile_apriori
    ):
        logger.critical(
            "No action asked, provide at least one of the following args:"
            "--sitelog, --modif_kw, --marker, --longname, --shortname, --station_info, --lfile_apriori"
        )
        return None

    # If force option provided, check if sitelog option too, if not, not relevant.
    if force_sitelog and not sitelog:
        logger.critical(
            "--force option is relevant only when --sitelog option with a **single** sitelog is also provided"
        )
        return None

    # If ignore option provided, check if sitelog option too, if not, not relevant.
    if ignore and not sitelog:
        logger.critical(
            "--ignore option is relevant only when using also --sitelog option"
        )
        return None

    if ninecharfile and not longname:
        logger.critical(
            "--ninecharfile option is relevant only when using also --longname option"
        )
        return None

    if (station_info and not lfile_apriori) or (not station_info and lfile_apriori):
        logger.critical("--station_info and --lfile_apriori must be provided together")
        return None

    if station_info and lfile_apriori and sitelog:
        logger.critical(
            "both sitelogs and GAMIT files given as metadata input. Managing both is not implemented yet"
        )
        return None

    # If inputfile doesn't exists, return
    if len(rinexinput) == 1 and not os.path.isfile(rinexinput[0]):
        logger.critical("The input file doesn't exist: %s", rinexinput)
        return None

    if len(rinexinput) > 1 and alone:
        logger.critical("several inputs are given while -a/--alone option is set")
        return None

    if output_logs and not os.path.isdir(output_logs):
        logger.critical(
            "The specified output folder for logs doesn't exist : " + output_logs
        )
        return None

    outputfolder = os.path.abspath(outputfolder)
    if not os.path.isdir(outputfolder):
        # mkdirs ???
        os.makedirs(outputfolder)

    # Creating log file
    now = datetime.now()
    nowstr = datetime.strftime(now, "%Y%m%d%H%M%S")

    if output_logs:
        logfolder = output_logs
    else:
        logfolder = outputfolder

    logfile = os.path.join(logfolder, nowstr + "_" + "rinexmod_errors.log")
    if verbose:
        _ = rimo_log.logger_define("DEBUG", logfile, "DEBUG")
    else:
        _ = rimo_log.logger_define("INFO", logfile, "INFO")

    ### the refactor of 2024-03 make this obsolete
    ## now rinexinput is ALWAYS a list
    # if isinstance(rinexinput, list):
    #     pass
    # elif alone:
    #     rinexinput = [rinexinput]
    ### the refactor of 2024-03 make this obsolete

    # Opening and reading lines of the file containing list of rinex to proceed
    if len(rinexinput) == 1 and not alone:
        try:
            rinexinput_use = [line.strip() for line in open(rinexinput[0]).readlines()]
        except:
            errmsg = f"Something went wrong while reading: {str(rinexinput[0])}"
            logger.error(errmsg)
            logger.error("Did you forget -a/--alone option?")
            return RinexModInputArgsError(errmsg)
    else:
        rinexinput_use = rinexinput

    if not rinexinput_use:
        errmsg = f"The input file is empty: {str(rinexinput)}"
        logger.error(errmsg)
        return RinexModInputArgsError(errmsg)

    if rinexinput_use[0].endswith("RINEX VERSION / TYPE"):
        errmsg = f"The input file is not a file list but a RINEX: {str(rinexinput[0])}"
        logger.error(errmsg)
        logger.error("Did you forget -a/--alone option?")
        return RinexModInputArgsError(errmsg)

    # sort the RINEX list
    if sort:
        rinexinput_use.sort()

    ### load the sitelogs/GAMIT-files as a **list of MetaData objects**
    # from sitelogs
    if sitelog:
        sitelogs_list_use = metadata_input_manage(sitelog, force_sitelog)
    # from GAMIT files
    elif station_info and lfile_apriori:
        sitelogs_list_use = gamit2mda_objs(
            station_info, lfile_apriori, force_fake_coords=force_fake_coords
        )
    else:
        sitelogs_list_use = None

    ### Looping in file list ###
    return_lists = dict()
    ####### Iterate over each RINEX
    rnxmod_kwargs_lis = []
    for rnx in rinexinput_use:
        rnxmod_kwargs = {
            "rinexfile": rnx,
            "outputfolder": outputfolder,
            "sitelog": sitelogs_list_use,
            "modif_kw": modif_kw,
            "marker": marker,
            "country": country,
            "longname": longname,
            "shortname": shortname,
            "force_rnx_load": force_rnx_load,
            "force_sitelog": force_sitelog,
            "ignore": ignore,
            "ninecharfile": ninecharfile,
            "no_hatanaka": no_hatanaka,
            "compression": compression,
            "relative": relative,
            "verbose": verbose,
            "return_lists": return_lists,
            "full_history": full_history,
            "filename_style": filename_style,
            "station_info": station_info,
            "lfile_apriori": lfile_apriori,
            "force_fake_coords": force_fake_coords,
            "remove": remove,
            "keep_rnx_rec": keep_rnx_rec,
        }

        rnxmod_kwargs_lis.append(rnxmod_kwargs)

    global rinexmod_mpwrap

    def rinexmod_mpwrap(rnxmod_kwargs_inp):
        try:
            return_lists_out = rinexmod(**rnxmod_kwargs_inp)
            return return_lists_out
        except Exception as e:
            if debug:  ### set as True for debug mode
                raise e
            else:
                logger.error(
                    "%s raised, RINEX is skiped: %s. Use -d/--debug option for more details",
                    type(e).__name__,
                    rnxmod_kwargs_inp["rinexfile"],
                )
                return None

    # number of parallel processing
    if multi_process > 1:
        logger.info("multiprocessing: %d cores used", multi_process)
    pool = mp.Pool(processes=multi_process)
    res_raw = [pool.apply_async(rinexmod_mpwrap, args=(x,)) for x in rnxmod_kwargs_lis]
    results = [e.get() for e in res_raw]

    for return_lists_mono in results:
        try:
            _return_lists_maker(return_lists_mono, return_lists)
        except ReturnListError:
            logger.warning(
                "one file has been skipped because of an odd individual return list"
            )
            continue

    #########################################
    logger.handlers.clear()

    if write:
        _return_lists_write(return_lists, logfolder, now)

    return return_lists


##### Class for --modif_kw
class ParseKwargs(argparse.Action):
    # source:
    # https://sumit-ghosh.com/posts/parsing-dictionary-key-value-pairs-kwargs-argparse-python/
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        icmt = 0
        for value in values:
            try:
                key, value = value.split("=")

                if key == "comment":
                    getattr(namespace, self.dest)[key + "_" + str(icmt)] = value
                    icmt += 1
                else:
                    getattr(namespace, self.dest)[key] = value

            except Exception as e:

                def _print_kw_tips(values_inp):
                    logger.critical("********************************************")
                    logger.critical("TIP1: be sure you have respected the syntax:")
                    logger.critical("      -k keyword1='value1' keyword2='value2'")
                    logger.critical("********************************************")
                    logger.critical(values_inp)
                    return None

                _print_kw_tips(values)
                raise e


# *****************************************************************************



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This script will get last version of sitelogs from M3G repository and write them
in an observatory dependent subfolder set in 'observatories'.
The -d --delete option will delete old version as to get only last version even
in a name changing case.
The -o --obsevatory option will download sitelogs for the specified observatory
only. Possible values : OVSM|OVSG|OVPF|REVOSIMA
USE :


* OUTPUTFOLDER : Folder where to write the downloaded sitelogs.

OPTION :

* -d : delete : Delete old sitelogs in storage folder. This allows to have only
                the last version, as version changing sitelogs changes of name.

* -m : move : Move old sitelogs into the given archive folder

* -o : observatory : IPGP's observatory ID that will be used to filter
                     sitelogs to download. Valid values are : OVSM|OVSG|OVPF|REVOSIMA.

* -r : root folder : per default, an observatory-specific folder is created
                     to store the corresponding sitelogs.
                     This option stores them in OUTPUTFOLDER's root

* -s : svn : a mode to maintain the legacy OVS's SVN folders.
             download the sitelog of a single obs and perform a 'svn commit'
             a single observatory must be given with -o option.
             the root folder option is automatically activated (-r)

* -f : force : Force download even if an identical sitelog already exists locally

* -e : exclude : Site(s) you want to exclude from download.
                 Provide as input 4 or 9 character site codes separated with spaces

EXAMPLE:

./get_m3g_sitelogs.py OUTPUTFOLDER (-d) (-o OVSM|OVSG|OVPF|REVOSIMA) (-r) (-s)

2021-10-18 FL
2023-03-21 PS
"""

import rinexmod.get_m3g


def main():
    import argparse

    # Parsing Args
    parser = argparse.ArgumentParser(
        description="Get the last versions of the IPGP sitelogs sorted on the M3G Database"
    )
    parser.add_argument(
        "sitelogsfolder",
        type=str,
        help="Output folder where to store downloaded sitelogs",
    )
    parser.add_argument(
        "-d",
        "--delete",
        help="Delete old sitelogs in the output folder. "
             "This allows to have only the last version, "
             "as version changing sitelogs changes of name.",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--move",
        help="Move old sitelogs into the given archive folder",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-o",
        "--observatory",
        nargs="+",
        help="Download sitelogs for some specific IPGP's observatories."
             "Valid values are : OVSM OVSG OVPF REVOSIMA OGA",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-r",
        "--root",
        help="Store the sitelogs in OUTPUTFOLDER root."
             "(per default, an observatory-specific folder is"
             "created to store the corresponding sitelogs.)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-s",
        "--svn",
        help="A mode to maintain the legacy OVS SVN folder."
             "Download the sitelog of a single obs and perform a svn commit."
             "A single observatory must be given with -o option."
             "The root folder option is automatically activated (-r)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-f",
        "--force",
        help="Force download even if an identical sitelog already exists locally",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-e",
        "--exclude",
        help="Site(s) you want to exclude from download. "
             "Provide as input 4 or 9 character site codes separated with spaces",
        nargs="+",
        default=[],
    )

    args = parser.parse_args()

    rinexmod.get_m3g.get_m3g_sitelogs(
        sitelogsfolder=args.sitelogsfolder,
        delete=args.delete,
        observatory=args.observatory,
        root_folder=args.root,
        svn_mode=args.svn,
        move_folder=args.move,
        force=args.force,
        exclude=args.exclude,
    )


if __name__ == "__main__":
    main()
